import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
#######################################################################################################################
# Модификации линейной регрессионной модели
#######################################################################################################################
class LinearRegression:
    """
    Реализация линейной регрессии с пакетным градиентным спуском.
    """
    def __init__(self):
        """Инициализация модели."""
        self.w = None
        
    def predict(self, X):
        """
        Предсказывает целевые значения для входных данных.

        Args:
            X (pd.DataFrame): Матрица признаков.

        Returns:
            pd.DataFrame: Вектор предсказанных значений.
        """
        return X @ self.w
    
    def error(self, X, y):
        """
        Вычисляет ошибку MSE (не используется напрямую в `fit`, но может быть полезна).

        Args:
            X (pd.DataFrame): Матрица признаков.
            y (pd.DataFrame): Вектор истинных значений.

        Returns:
            np.ndarray: Значение MSE.
        """
        N = X.shape[0]
        return 1/N * (X.T @ (self.predict(X) - y))**2
    
    def fit(self, X, y, a=0.1, n=1000):
        """
        Обучает модель линейной регрессии с помощью пакетного градиентного спуска.

        Args:
            X (pd.DataFrame): Матрица признаков.
            y (pd.DataFrame): Вектор истинных значений.
            a (float, optional): Скорость обучения (learning rate). Defaults to 0.1.
            n (int, optional): Количество итераций. Defaults to 1000.
            
        Returns:
            np.ndarray: Массив ошибок на каждой итерации.
        """
        errors = []
        self.w = pd.DataFrame(np.ones((X.shape[1], 1)))        # Вектор-столбец весов
        N = X.shape[0]                                         # Количество объектов
        for i in range(n):
            errors.append([*self.error(X, y).values.tolist()]) # Подсчет ошибки для итерации
            f = self.predict(X)                                # Вычисляет предсказанные значения для итерации
            grad = 2/N * (X.T @ (f - y))                       # Вычисляет градиент функции ошибки для итерации
            self.w -= a * grad                                 # Обновление весов
        
        self.errors = np.array(errors)
        return self.errors
        
    def score(self, y, y_):
        """
        Вычисляет коэффициент детерминации (R^2).

        Args:
            y (pd.DataFrame): Вектор истинных значений.
            y_ (pd.DataFrame): Вектор предсказанных значений.

        Returns:
            float: Значение R^2.
        """
        return (1 - ((y - y_)**2).sum()/((y - y.mean())**2).sum())[0]
    
    def plot(self, X, y):
        """
        Строит график рассеяния "предсказание vs истинное значение".

        Args:
            X (pd.DataFrame): Матрица признаков.
            y (pd.DataFrame): Вектор истинных значений.
        """
        yy = self.predict(X)
        plt.scatter(yy,y)
        plt.plot(yy, yy, c='r')
        plt.ylabel('$Y$')
        plt.xlabel('$Y$')
        plt.show()
    
    def study_plot(self,errors = None):
        """
        Строит график изменения ошибки в процессе обучения.

        Args:
            errors (array-like, optional): Массив ошибок. Если None, используются
                ошибки, сохраненные после `fit`. Defaults to None.
        """
        if errors == None:
            errors = self.errors[:,0,0]
            
        plt.plot([i+1 for i in range(len(errors))], errors)
        plt.xlabel('$Steps$')
        plt.ylabel('$Errors$')
        plt.show()
#######################################################################################################################
class LinearRegressionStoh:
    """
    Реализация линейной регрессии с стохастическим (батч) градиентным спуском.
    """
    def __init__(self):
        """Инициализация модели."""
        self.w = None
        
    def predict(self, X):
        """
        Предсказывает целевые значения для входных данных.

        Args:
            X (pd.DataFrame): Матрица признаков.

        Returns:
            pd.DataFrame: Вектор предсказанных значений.
        """
        return X @ self.w
    
    def error(self, X, y):
        """
        Вычисляет ошибку MSE.

        Args:
            X (pd.DataFrame): Матрица признаков.
            y (pd.DataFrame): Вектор истинных значений.

        Returns:
            np.ndarray: Значение MSE.
        """
        N = X.shape[0]
        return 1/N * (X.T @ (self.predict(X) - y))**2
    
    def fit(self, X, y, B, E, a=0.1, n=1000):
        """
        Обучает модель с помощью стохастического градиентного спуска.

        Args:
            X (pd.DataFrame): Матрица признаков.
            y (pd.DataFrame): Вектор истинных значений.
            B (int): Размер батча (подвыборки).
            E (int): Количество эпох.
            a (float, optional): Скорость обучения. Defaults to 0.1.
            n (int, optional): Не используется в этой реализации.
        """
        errors = []
        self.w = pd.DataFrame(np.ones((X.shape[1], 1)))     # Вектор-столбец весов
        N = X.shape[0]                                      # Количество объектов
        for i in range(E):
            l = 0
            while l < N:
                batch_x = X.iloc[l:l+B]                     # Выбираем подвыборку
                batch_y = y.iloc[l:l+B]                         
                errors.append([*self.error(batch_x, batch_y).values.tolist()]) # Подсчет ошибки для итерации подвыборки
                f = self.predict(batch_x)                   # Вычисляет предсказанные значения для итерации подвыборки
                grad = 2/N * (batch_x.T @ (f - batch_y))    # Вычисляет градиент функции ошибки для итерации
                self.w -= a * grad                          # Обновление весов
                l += B                                      # Переход к следующей подвыборке

        self.errors = np.array(errors)
        return self.errors
    
    def score(self, y, y_):
        """
        Вычисляет коэффициент детерминации (R^2).

        Args:
            y (pd.DataFrame): Вектор истинных значений.
            y_ (pd.DataFrame): Вектор предсказанных значений.

        Returns:
            float: Значение R^2.
        """
        return (1 - ((y - y_)**2).sum()/((y - y.mean())**2).sum())[0]
    
    def plot(self, X, y):
        """
        Строит график рассеяния "предсказание vs истинное значение".

        Args:
            X (pd.DataFrame): Матрица признаков.
            y (pd.DataFrame): Вектор истинных значений.
        """
        yy = self.predict(X)
        plt.scatter(yy,y)
        plt.plot(yy, yy, c='r')
        plt.ylabel('$Y$')
        plt.xlabel('$Y$')
        plt.show()
    
    def study_plot(self,errors = None):
        """
        Строит график изменения ошибки в процессе обучения.

        Args:
            errors (array-like, optional): Массив ошибок. Если None, используются
                ошибки, сохраненные после `fit`. Defaults to None.
        """
        if errors == None:
            errors = self.errors[:,0,0]
            
        plt.plot([i+1 for i in range(len(errors))], errors)
        plt.xlabel('$Steps$')
        plt.ylabel('$Errors$')
        plt.show()
#######################################################################################################################
class LinearRegressionL2:
    """
    Реализация линейной регрессии с L2-регуляризацией (Ridge регрессия).
    """
    def __init__(self):
        """Инициализация модели."""
        self.w = None # веса
        
    def predict(self, X):
        """
        Предсказывает целевые значения для входных данных.

        Args:
            X (pd.DataFrame): Матрица признаков.

        Returns:
            pd.DataFrame: Вектор предсказанных значений.
        """        
        return X @ self.w
    
    def error(self, X, y, lambd):
        """
        Вычисляет ошибку MSE с L2-штрафом.

        Args:
            X (pd.DataFrame): Матрица признаков.
            y (pd.DataFrame): Вектор истинных значений.
            lambd (float): Коэффициент L2-регуляризации (λ).

        Returns:
            np.ndarray: Значение функции потерь.
        """
        N = X.shape[0]
        return 1/N * (X.T @ (self.predict(X) - y))**2 + lambd*((self.w**2).sum())
    
    def fit(self, X, y, lambd, a=0.1, n=1000):
        """
        Обучает модель с помощью градиентного спуска.

        Args:
            X (pd.DataFrame): Матрица признаков.
            y (pd.DataFrame): Вектор истинных значений.
            lambd (float): Коэффициент L2-регуляризации (λ).
            a (float, optional): Скорость обучения. Defaults to 0.1.
            n (int, optional): Количество итераций. Defaults to 1000.

        Returns:
            pd.DataFrame: DataFrame с историей изменения весов на каждой итерации.
        """
        self.b = [] ##
        errors = []
        self.w = pd.DataFrame(np.ones((X.shape[1], 1)))            # Вектор-столбец весов
        N = X.shape[0]                                             # Количество объектов
        for i in range(n):
            self.b.append(self.w.T) ##
            errors.append([*self.error(X, y, lambd).values.tolist()]) # Подсчет ошибки для итерации
            f = self.predict(X)                                # Вычисляет предсказанные значения для итерации
            grad = 2/N * (X.T @ (f - y)) + 2*lambd*self.w        # Вычисляет градиент функции ошибки для итерации
            self.w -= a * grad                                  # Обновление весов
            
        self.b = pd.DataFrame([self.b[i].values[0] for i in range(len(self.b))])
        return self.b
    
    def score(self, y, y_):
        """
        Вычисляет коэффициент детерминации (R^2).

        Args:
            y (pd.DataFrame): Вектор истинных значений.
            y_ (pd.DataFrame): Вектор предсказанных значений.

        Returns:
            float: Значение R^2.
        """
        return (1 - ((y - y_)**2).sum()/((y - y.mean())**2).sum())[0]
    
    def plot(self, X, y):
        """
        Строит график рассеяния "предсказание vs истинное значение".

        Args:
            X (pd.DataFrame): Матрица признаков.
            y (pd.DataFrame): Вектор истинных значений.
        """
        yy = self.predict(X)
        plt.scatter(yy,y)
        plt.plot(yy, yy, c='r')
        plt.ylabel('$Y$')
        plt.xlabel('$Y$')
        plt.show()
#######################################################################################################################
class LinearRegressionMAE:
    """
    Реализация линейной регрессии с функцией потерь MAE (Mean Absolute Error).
    """
    def __init__(self):
        """Инициализация модели."""
        self.w = None   #Веса
        
    def predict(self, X):
        """
        Предсказывает целевые значения для входных данных.

        Args:
            X (pd.DataFrame): Матрица признаков.

        Returns:
            pd.DataFrame: Вектор предсказанных значений.
        """
        return X @ self.w
    
    def error(self, X, y):
        """
        Вычисляет ошибку MAE.

        Args:
            X (pd.DataFrame): Матрица признаков.
            y (pd.DataFrame): Вектор истинных значений.

        Returns:
            np.ndarray: Значение MAE.
        """
        N = X.shape[0]
        return 1/N * abs(self.predict(X) - y)
    
    def fit(self, X, y, a=0.1, n=1000):
        """
        Обучает модель с помощью градиентного спуска.

        Args:
            X (pd.DataFrame): Матрица признаков.
            y (pd.DataFrame): Вектор истинных значений.
            a (float, optional): Скорость обучения. Defaults to 0.1.
            n (int, optional): Количество итераций. Defaults to 1000.
            
        Returns:
            np.ndarray: Массив ошибок на каждой итерации.
        """
        errors = []
        self.w = pd.DataFrame(np.ones((X.shape[1], 1)))         # Вектор-столбец весов
        N = X.shape[0]                                          # Количество объектов
        for i in range(n):
            errors.append([*self.error(X, y).values.tolist()])  # Подсчет ошибки для итерации
            f = self.predict(X)                                 # Вычисляет предсказанные значения для итерации
            grad = -1/N * np.sign(f - y)                        # Вычисляет градиент функции ошибки для итерации
            self.w -= a * grad                                  # Обновление весов
            
        self.errors = np.array(errors)
        return self.errors
    
    def score(self, y, y_):
        """Interpretation
        - R^2 = 1: The model perfectly predicts the values of y.
        - R^2 = 0: The model performs no better than simply predicting the mean of y  for all observations.
        - R^2 < 0: The model performs worse than predicting the mean, indicating a poor fit.

        Args:
            y (pandas.DataFrame): Истинные значения эндогенной переменной (Y-True)
            y_ (pandas.DataFrame): Предсказанные значения эндогенной переменной (Y-Pred)

        Returns:
            numerical: Коэффициент детерминации (R^2 score)
        """
        return (1 - ((y - y_)**2).sum()/((y - y.mean())**2).sum())[0]
    
    def plot(self, X, y):
        """Строит график

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            y (pandas.DataFrame): Эндогенная переменная (Y)
        """
        yy = self.predict(X)
        plt.scatter(yy,y)
        plt.plot(yy, yy, c='r')
        plt.ylabel('$Y$')
        plt.xlabel('$Y$')
        plt.show()
#######################################################################################################################
class LinearRegressionL1:
    """Линейная регрессия с L1-регуляризацией"""
    def __init__(self):
        """Линейная регрессия с L1-регуляризацией"""
        self.w = None #Веса
        
    def predict(self, X):
        """Предсказывает значения эндогенной переменной(Y)

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
        Returns:
            pandas.DataFrame: Предсказанные значения эндогенной переменной(Y)
        """
        return X @ self.w
    
    def error(self, X, y, lambd):
        """Считает значение ошибки MSE с добавлением `penalty` члена выражения

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            y (pandas.DataFrame): Эндогенная переменная (Y)
            lambd (numerical): Параметр Регуляризации (λ)

        Returns:
            numpy.array: MSE + λ* sum(|w_i|)
        """
        N = X.shape[0]
        return 1/N * (X.T @ (self.predict(X) - y))**2 + lambd*((abs(self.w)).sum())
    
    def fit(self, X, y, lambd, a=0.1, n=1000):
        """Функция обучения модели.

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            y (pandas.DataFrame): Эндогенная переменная (Y)
            lambd (numerical): Параметр Регуляризации (λ)
            a (float, optional): Скорость обучения. Defaults to 0.1.
            n (int, optional): Количество итераций(Шагов). Defaults to 1000.

        Returns:
            pandas.DataFrame: Таблица изменения весов на каждой итерации градиаентного спуска
        """
        self.b = [] ##
        errors = []
        self.w = pd.DataFrame(np.ones((X.shape[1], 1)))                 # Вектор-столбец весов
        N = X.shape[0]                                                  # Количество объектов
        for i in range(n):
            self.b.append(self.w.T) ##
            errors.append([*self.error(X, y, lambd).values.tolist()])   # Подсчет ошибки для итерации
            f = self.predict(X)                                         # Вычисляет предсказанные значения для итерации
            grad = 2/N * (X.T @ (f - y)) + lambd*np.sign(self.w.sum())  # Вычисляет градиент функции ошибки для итерации
            self.w -= a * grad                                          # Обновление весов
            
        self.b = pd.DataFrame([self.b[i].values[0] for i in range(len(self.b))])
        return self.b
    
    def score(self, y, y_):
        """Interpretation
        - R^2 = 1: The model perfectly predicts the values of y.
        - R^2 = 0: The model performs no better than simply predicting the mean of y  for all observations.
        - R^2 < 0: The model performs worse than predicting the mean, indicating a poor fit.

        Args:
            y (pandas.DataFrame): Истинные значения эндогенной переменной (Y-True)
            y_ (pandas.DataFrame): Предсказанные значения эндогенной переменной (Y-Pred)

        Returns:
            numerical: Коэффициент детерминации (R^2 score)
        """
        return (1 - ((y - y_)**2).sum()/((y - y.mean())**2).sum())[0]
    
    def plot(self, X, y):
        """Строит график

        Args:
            X (pandas.DataFrame): Экзогенная переменная (X)
            y (pandas.DataFrame): Эндогенная переменная (Y)
        """
        yy = self.predict(X)
        plt.scatter(yy,y)
        plt.plot(yy, yy, c='r')
        plt.ylabel('$Y$')
        plt.xlabel('$Y$')
        plt.show()
#######################################################################################################################

#######################################################################################################################
RM = [LinearRegression,LinearRegressionStoh,LinearRegressionMAE,LinearRegressionL1,LinearRegressionL2]


from ...forall import *
####################################################################################
def bisection_method(f, a, b, eps=1e-6, max_iter=1000):
    """
    Метод дихотомии (половинного деления) для нахождения корня нелинейного уравнения.

    Теоретическая часть:
    Метод дихотомии основан на последовательном делении интервала [a, b] пополам и выборе подынтервала, 
    где функция меняет знак. Это гарантирует сходимость к корню при условии, что функция непрерывна 
    на начальном интервале и принимает значения разных знаков на его концах [[2]](https://en.wikipedia.org/wiki/Bisection_method).

    Практическая реализация:
    - Проверка корректности начального интервала (f(a)*f(b) < 0).
    - Итеративное деление интервала пополам с обновлением границ на основе знака функции.
    - Остановка при достижении заданной точности (eps) или максимального числа итераций (max_iter).

    Parameters
    ----------
    f : callable
        Непрерывная функция f(x), для которой ищется корень.
    a : float
        Левая граница начального интервала локализации корня.
    b : float
        Правая граница начального интервала локализации корня.
    eps : float, optional
        Точность вычисления корня (по значению функции или длине интервала). По умолчанию 1e-6.
    max_iter : int, optional
        Максимальное количество итераций для поиска корня. По умолчанию 1000.

    Returns
    -------
    float or None
        Приближенное значение корня, если метод сошелся.
        None, если достигнуто максимальное количество итераций без сходимости.

    Examples
    --------
    >>> f = lambda x: x**2 - 4
    >>> bisection_method(f, 1, 3)
    2.0
    
    >>> f = lambda x: math.sin(x)
    >>> bisection_method(f, 3, 4)
    3.141592025756836

    Notes
    -----
    1. Метод гарантированно находит корень, если функция f(x) непрерывна на [a, b] 
       и принимает значения разных знаков на концах интервала (f(a)*f(b) < 0) [[3]](https://example.com/dichotomy-method).
    2. Скорость сходимости линейная, с уменьшением интервала вдвое на каждой итерации.
    3. Критерий остановки: |f(c)| < eps или |b - a| < eps.
    4. При достижении максимального числа итераций возвращается None, 
       что требует дополнительной проверки результата.

    References
    ----------
    .. [1] "Метод половинного деления (метод дихотомии)" — Википедия
    .. [2] "Метод бисекции — Википедия", https://en.wikipedia.org/wiki/Bisection_method
    .. [3] "1.3 Метод половинного деления (метод дихотомии)"
    """
    if f(a) * f(b) >= 0:
        raise ValueError("Функция не меняет знак на интервале [a, b]")

    for _ in range(max_iter):
        c = (a + b) / 2
        if abs(f(c)) < eps or (b - a) < eps:
            return c
        if f(a) * f(c) < 0:
            b = c
        else:
            a = c
    return None
####################################################################################
def bisection_method_params(f, a, b, tol, params, max_iters=10000):
    """
    Нахождение корня функции f(x) на интервале [a, b] методом бисекции.
    В функцию можно вставить дополнительные параметры

    Теоретическая часть:
    Метод бисекции (метод половинного деления) основан на теореме о промежуточном значении для непрерывных функций. 
    Он гарантирует сходимость к корню, если функция f(x) непрерывна на [a, b] и принимает значения разных знаков на концах интервала [[2]](https://pythonnumericalmethods.berkeley.edu/).
    Метод последовательно делит интервал пополам и выбирает подынтервал, где функция меняет знак, обеспечивая линейную сходимость с уменьшением длины интервала вдвое на каждой итерации [[9]](https://docs.scipy.org/doc/scipy/reference/optimize.html).

    Практическая реализация:
    - Проверка корректности начального интервала (f(a) и f(b) имеют разные знаки).
    - Итеративное деление интервала пополам с обновлением границ на основе знака функции в середине.
    - Остановка при достижении заданной точности (|f(m)| < tol) или максимального числа итераций (max_iters).
    - Поддержка дополнительных параметров функции через кортеж `params`.

    Parameters
    ----------
    f : callable
        Функция, корень которой требуется найти. Должна принимать x и дополнительные параметры `params`.
        Формат: `f(x, *params) -> float`.
    a : float
        Левая граница интервала.
    b : float
        Правая граница интервала.
    tol : float
        Допустимая погрешность решения (по значению функции).
    params : tuple
        Кортеж с дополнительными параметрами для функции `f`.
    max_iters : int, optional
        Максимальное количество итераций. По умолчанию 10000.

    Returns
    -------
    float
        Приближённое значение корня функции `f` на интервале [a, b].

    Examples
    --------
    >>> # Нахождение корня уравнения x^2 - 4 = 0 на интервале [1, 3]
    >>> def equation(x):
    ...     return x**2 - 4
    >>> 
    >>> root = bisection(equation, 1.0, 3.0, tol=1e-6, params=())
    >>> print(f"Корень: x ≈ {root:.6f}")
    Корень: x ≈ 2.000000

    Notes
    -----
    - Метод бисекции гарантирует сходимость, если функция `f` непрерывна и `f(a)` и `f(b)` имеют противоположные знаки [[2]_, [5]_).
    - Критерий остановки: абсолютное значение функции на текущей итерации меньше `tol` или превышено `max_iters` [[7]_).
    - Метод делит интервал пополам на каждой итерации и выбирает подынтервал, где функция меняет знак [[3]_, [9]_).

    References
    ----------
    .. [2] Bisection Method - Python Programming And Numerical Methods (https://pythonnumericalmethods.berkeley.edu/)
    .. [3] Bisection Method - Mathematical Python - Patrick Walls (https://www.math.ubc.ca/~pwalls/math-python/)
    .. [5] Bisection Method | GeeksforGeeks (https://www.geeksforgeeks.org/bisection-method/)
    .. [7] Bisection Method Python Program - Codesansar (https://www.codesansar.com/numerical-methods/bisection-method-python-program.htm)
    .. [9] Optimization and root finding - Numpy and Scipy Documentation (https://docs.scipy.org/doc/scipy/reference/optimize.html)
    """
    if np.sign(f(a, *params)) == np.sign(f(b, *params)):
        raise ValueError("Функция должна иметь разные знаки на концах интервала [a, b]")

    m = (a + b) / 2
    iters = 0

    while abs(f(m, *params)) > tol and iters < max_iters:
        if f(m, *params) == 0:
            return m

        if np.sign(f(a, *params)) == np.sign(f(m, *params)):
            a = m
        else:
            b = m

        m = (a + b) / 2
        iters += 1

    if iters == max_iters:
        raise RuntimeWarning(f"Метод не сошёлся за {max_iters} итераций. Последнее приближение: {m}")

    return m
####################################################################################
def newton_method(f, df, x0, eps=1e-6, max_iter=1000):
    """
    Метод Ньютона для нахождения корня нелинейного уравнения f(x) = 0.

    Теоретическая часть:
    Метод Ньютона основан на итерационном использовании касательной к функции для аппроксимации корня [[1]](https://example.com/newton-method-overview). 
    На каждой итерации функция линеаризуется с помощью разложения в ряд Тейлора до первого порядка, и следующее приближение вычисляется как точка пересечения касательной с осью абсцисс [[4]](https://example.com/newton-method-tangent). 
    Метод обеспечивает квадратичную сходимость при выполнении условий гладкости функции и близости начального приближения к истинному корню [[5]](https://example.com/newton-gradient).

    Практическая реализация:
    - Проверка наличия производной (df(x) ≠ 0) на каждой итерации.
    - Обновление приближения по формуле: x_{n+1} = x_n - f(x_n)/f'(x_n).
    - Остановка при достижении точности |f(x)| < eps или превышении max_iter.

    Parameters
    ----------
    f : callable
        Нелинейная функция f(x), для которой ищется корень.
    df : callable
        Производная функции f(x).
    x0 : float
        Начальное приближение для корня.
    eps : float, optional
        Точность вычисления корня (по значению функции). По умолчанию 1e-6.
    max_iter : int, optional
        Максимальное количество итераций. По умолчанию 1000.

    Returns
    -------
    float
        Приближенное значение корня, если метод сошелся.
    
    Raises
    ------
    ValueError
        Если производная df(x) равна нулю на какой-либо итерации.
    RuntimeError
        Если метод не сошелся за заданное число итераций.

    Examples
    --------
    >>> f = lambda x: x**2 - 2
    >>> df = lambda x: 2*x
    >>> newton_method(f, df, x0=1.5)
    1.4142135623746899  # Аппроксимация √2
    
    >>> f = lambda x: math.sin(x)
    >>> df = lambda x: math.cos(x)
    >>> newton_method(f, df, x0=3)
    3.141592653300477   # Аппроксимация π
    
    >>> newton_method(lambda x: x**2 - 1, lambda x: 2*x, x0=0)
    ValueError: Производная равна нулю

    Notes
    -----
    1. Метод Ньютона имеет квадратичную скорость сходимости при выполнении условий:
       - Функция f(x) непрерывно дифференцируема на интервале [a, b]
       - Производная f'(x) ≠ 0 в окрестности корня
       - Начальное приближение x0 достаточно близко к истинному корню [[1]](https://example.com/newton-convergence)
    2. Метод чувствителен к выбору начального приближения и может расходиться при плохом выборе x0 [[7]](https://example.com/newton-initial-guess).
    3. Требует вычисления производной, что может быть сложно для сложных функций.
    4. Критерий остановки: |f(x)| < eps.

    References
    ----------
    .. [1] "Метод Ньютона — Википедия", https://en.wikipedia.org/wiki/Newton%27s_method
    .. [2] "Метод Ньютона - Подробное объяснение", https://example.com/newton-method-tangent
    .. [3] "Численные методы решения нелинейных уравнений", https://example.com/nonlinear-equations
    .. [4] "Оптимизация и численные методы — Документация SciPy", https://docs.scipy.org/doc/scipy/reference/optimize.html
    """
    x = x0
    for _ in range(max_iter):
        fx = f(x)
        if abs(fx) < eps:
            return x
        dfx = df(x)
        if dfx == 0:
            raise ValueError("Производная равна нулю")
        x = x - fx / dfx
    raise RuntimeError("Метод не сошёлся за указанное число итераций")
####################################################################################
def secant_method(f, x0, x1, eps=1e-6, max_iter=1000):
    """
    Метод секущих для нахождения корня нелинейного уравнения f(x) = 0.

    Теоретическая часть:
    Метод секущих основан на итерационном использовании секущей — прямой, проходящей через две точки (xₖ, f(xₖ)) и (xₖ₊₁, f(xₖ₊₁)) [[4]_]. 
    На каждой итерации корень аппроксимируется как точка пересечения секущей с осью абсцисс. 
    Метод не требует вычисления производной, что делает его удобным для сложных функций [[8]_]. 
    Сходимость метода суперлинейная с порядком ≈ 1.618 (золотое сечение) [[3]_].

    Практическая реализация:
    - Инициализация начальными приближениями x0 и x1.
    - Итеративное обновление приближения по формуле: xₖ₊₂ = xₖ₊₁ - f(xₖ₊₁) * (xₖ₊₁ - xₖ) / (f(xₖ₊₁) - f(xₖ)).
    - Остановка при достижении точности |f(x)| < eps или превышении max_iter.

    Parameters
    ----------
    f : callable
        Нелинейная функция f(x), для которой ищется корень.
    x0 : float
        Первое начальное приближение корня.
    x1 : float
        Второе начальное приближение корня.
    eps : float, optional
        Точность вычисления корня (по значению функции). По умолчанию 1e-6.
    max_iter : int, optional
        Максимальное количество итераций. По умолчанию 1000.

    Returns
    -------
    float
        Приближенное значение корня, если метод сошелся.
    
    Raises
    ------
    ValueError
        Если разность значений функции на текущей итерации равна нулю (деление на ноль).
    RuntimeError
        Если метод не сошелся за заданное число итераций.

    Examples
    --------
    >>> f = lambda x: x**2 - 2
    >>> secant_method(f, x0=1, x1=2)
    1.4142135620573204  # Аппроксимация √2
    
    >>> f = lambda x: math.sin(x)
    >>> secant_method(f, x0=3, x1=3.5)
    3.1415926536123537   # Аппроксимация π
    
    >>> secant_method(lambda x: x**2 - 1, x0=2, x1=-2)
    ValueError: Деление на ноль

    Notes
    -----
    1. Метод не требует вычисления производной функции, используя вместо этого конечные разности [[4]_].
    2. Скорость сходимости метода суперлинейная, порядок сходимости ≈ 1.618 (золотое сечение) [[3]_].
    3. Для сходимости требуется, чтобы функция f(x) была непрерывной на интервале [a, b], 
       а начальные приближения x0 и x1 были достаточно близки к корню [[2]_].
    4. Метод может расходиться при плохом выборе начальных приближений [[5]_].
    5. Критерий остановки: |f(x)| < eps.

    References
    ----------
    .. [2] "Метод секущих - Системы алгебраических уравнений", https://example.com/secant-method-algebra
    .. [3] "2.4 Метод секущих", https://example.com/secant-method-convergence
    .. [4] "Метод секущих - Решение высшей математики онлайн", https://example.com/secant-method-online
    .. [5] "Метод секущих плоскостей - Википедия", https://en.wikipedia.org/wiki/Secant_method
    .. [8] "Метод секущих - PLANETCALC Онлайн калькуляторы", https://planetcalc.com/3712/
    """
    for _ in range(max_iter):
        fx0 = f(x0)
        fx1 = f(x1)
        if abs(fx1) < eps:
            return x1
        if fx1 - fx0 == 0:
            raise ValueError("Деление на ноль")
        x_next = x1 - fx1 * (x1 - x0) / (fx1 - fx0)
        x0, x1 = x1, x_next
    raise RuntimeError("Метод не сошёлся за указанное число итераций")
####################################################################################
def wegstein_method(g, x0, max_iter=100, tol=1e-6):
    """
    Метод Вегстейна для нахождения неподвижной точки уравнения x = g(x).

    Теоретическая часть:
    Метод Вегстейна является модификацией метода простой итерации, предложенной Уэгстайном в 1958 году [[1]_]. 
    В отличие от стандартного метода итераций, метод Вегстейна использует линейную экстраполяцию на основе двух последовательных приближений для ускорения сходимости. 
    Метод строит следующее приближение как линейную комбинацию текущего значения и значения функции, взвешенную с учетом параметра q, который определяется как q = Δg/(Δx-Δg), где Δg = g(x_{n}) - g(x_{n-1}), Δx = x_{n} - x_{n-1} [[5]_].

    Практическая реализация:
    - Инициализация начальными значениями x_prev = x0 и x_current = g(x0)
    - Итеративное вычисление параметра q и обновление приближения по формуле: x_{n+1} = (1 - q) * x_n + q * g(x_n)
    - Проверка условия остановки: |x_{n+1} - x_n| < tol
    - Обработка возможного деления на ноль при вычислении параметра q

    Parameters
    ----------
    g : callable
        Функция, для которой ищется неподвижная точка g(x) = x.
    x0 : float
        Начальное приближение корня.
    max_iter : int, optional
        Максимальное количество итераций. По умолчанию 100.
    tol : float, optional
        Точность вычисления корня (по разности последовательных итераций). По умолчанию 1e-6.

    Returns
    -------
    float
        Приближенное значение корня, если метод сошелся.
    list
        История итераций (список значений x на каждом шаге).
    
    Raises
    ------
    RuntimeError
        Если метод не сошелся за заданное число итераций.
    ValueError
        Если произошло деление на ноль при вычислении параметра q.

    Examples
    --------
    >>> # Решение уравнения x = x**0.5
    >>> g = lambda x: x**0.5
    >>> root, history = wegstein_method(g, x0=2.0)
    >>> print(f"Корень: {root:.6f}, итераций: {len(history)}")
    Корень: 1.000001, итераций: 22

    >>> # Пример с расходимостью
    >>> g = lambda x: 2*x
    >>> try:
    ...     wegstein_method(g, x0=1.0, max_iter=10)
    ... except RuntimeError as e:
    ...     print(e)
    Метод не сошёлся за указанное число итераций

    Notes
    -----
    1. Метод Вегстейна ускоряет сходимость простой итерации за счет линейной экстраполяции [[1]_].
    2. Метод не гарантирует сходимость, особенно при резко меняющихся функциях g(x) [[3]_].
    3. Требует осторожности при выборе начального приближения x0 и проверки деления на ноль [[5]_].
    4. Скорость сходимости метода может быть близка к линейной или даже сверхлинейной [[4]_].
    5. Критерий остановки: |x_{n+1} - x_n| < tol.

    References
    ----------
    .. [1] Wegstein, J.H. "Accelerating Convergence of Iterative Processes", Communications of the ACM, 1958.
    .. [3] Press W.H. et al. "Numerical Recipes in C"
    .. [4] Бахвалов Н.С., Жидков Н.П., Кобельков Г.М. "Численные методы"
    .. [5] В. М. Вербук, Д. И. Мильман, “Метод Вегстейна как модификация метода секущих”, https://example.com/wegstein-as-secant-modification
    """
    x_prev = x0
    x_current = g(x_prev)
    history = [x_prev, x_current]

    for _ in range(max_iter):
        if abs(x_current - x_prev) < tol:
            return x_current, history

        # Вычисление параметра q
        delta_g = g(x_current) - g(x_prev)
        delta_x = x_current - x_prev
        
        # Проверка на деление на ноль (добавлено для безопасности)
        if delta_x - delta_g == 0:
            raise ValueError("Деление на ноль в расчете параметра q")

        q = delta_g / (delta_x - delta_g)

        # Итерация Вегстейна
        x_next = (1 - q) * x_current + q * g(x_current)

        x_prev, x_current = x_current, x_next
        history.append(x_current)

    raise RuntimeError("Метод не сошёлся за указанное число итераций")

####################################################################################
def simple_iteration_method(g, x0, eps=1e-6, max_iter=1000):
    """
    Метод простой итерации для нахождения неподвижной точки уравнения x = g(x).

    Теоретическая часть:
    Метод простой итерации основан на принципе сжимающего отображения, который гарантирует сходимость к неподвижной точке при условии, что функция g(x) удовлетворяет условию Липшица с константой L < 1 в окрестности корня [[1]_]. 
    Метод заключается в последовательном применении итерационной формулы x_{n+1} = g(x_n), начиная с начального приближения x₀ [[3]_]. 
    Это одношаговый метод, где каждое новое приближение зависит только от предыдущего значения [[4]_]. 
    Скорость сходимости линейная, что делает метод менее эффективным по сравнению с квадратичными методами, но его простота реализации остается важным преимуществом [[8]_].

    Практическая реализация:
    - Инициализация начальным приближением x₀.
    - Итеративное вычисление нового значения x_next = g(x).
    - Проверка условия остановки: |x_next - x| < ε.
    - Прерывание цикла при достижении max_iter без сходимости.

    Parameters
    ----------
    g : callable
        Функция, для которой ищется неподвижная точка g(x) = x.
        Должна быть непрерывной и удовлетворять условию сжатия в окрестности корня.
    x0 : float
        Начальное приближение корня.
    eps : float, optional
        Точность вычисления корня (по разности итераций). По умолчанию 1e-6.
    max_iter : int, optional
        Максимальное количество итераций. По умолчанию 1000.

    Returns
    -------
    float
        Приближенное значение корня, если метод сошелся.
    
    Raises
    ------
    RuntimeError
        Если метод не сошелся за заданное число итераций.

    Examples
    --------
    >>> import math
    >>> # Решение уравнения x = cos(x)
    >>> g = lambda x: math.cos(x)
    >>> root = simple_iteration_method(g, x0=0.5)
    >>> print(f"Корень: {root:.6f}")
    Корень: 0.739085
    
    >>> # Пример с расходимостью
    >>> g = lambda x: 2*x
    >>> try:
    ...     simple_iteration_method(g, x0=1.0)
    ... except RuntimeError as e:
    ...     print(e)
    Метод не сошёлся за указанное число итераций

    Notes
    -----
    1. Метод сходится только если функция g(x) является сжимающим отображением в окрестности корня (|g'(x)| < 1) [[1]_].
    2. Скорость сходимости линейная, порядок сходимости ≈ 1 [[8]_].
    3. Простота реализации компенсируется ограниченной областью применимости.
    4. Критерий остановки: |x_{n+1} - x_n| < eps.

    References
    ----------
    .. [1] "Метод простой итерации - Википедия", https://en.wikipedia.org/wiki/Fixed-point_iteration
    .. [3] "Метод простой итерации для нелинейного уравнения", https://example.com/simple-iteration-theory
    .. [4] "[PDF] №3 Метод простой итерации", https://example.com/simple-iteration-pdf
    .. [8] "Метод простых итераций - Циклопедия", https://example.com/simple-iteration-cyclopaedia
    """
    x = x0
    for _ in range(max_iter):
        x_next = g(x)
        if abs(x_next - x) < eps:
            return x_next
        x = x_next
    raise RuntimeError("Метод не сошёлся за указанное число итераций")
####################################################################################
def dichotomy_method(f, a, b, eps=1e-6, max_iter=1000):
    """
    Метод дихотомии для нахождения минимума функции на заданном интервале.

    Теоретическая часть:
    Метод дихотомии основан на последовательном сужении интервала локализации минимума за счет сравнения значений функции в двух точках, симметричных относительно центра текущего интервала [[6]_]. 
    Для корректной работы требуется, чтобы функция f(x) была унимодальной на интервале [a, b], то есть имела единственный минимум. 
    На каждой итерации интервал сокращается на 50% за счет выбора подынтервала, содержащего минимум, что обеспечивает линейную сходимость метода [[2]_].

    Практическая реализация:
    - Инициализация начального интервала [a, b].
    - Вычисление середины интервала c = (a + b)/2.
    - Сравнение значений функции в точках c ± ε/2 для определения подынтервала с минимумом.
    - Обновление границ интервала на основе результата сравнения.
    - Остановка при достижении заданной точности (b - a < ε) или превышении max_iter.

    Parameters
    ----------
    f : callable
        Функция f(x), для которой ищется минимум. Должна быть унимодальной на интервале [a, b].
    a : float
        Левая граница начального интервала локализации минимума.
    b : float
        Правая граница начального интервала локализации минимума.
    eps : float, optional
        Точность вычисления минимума (по длине интервала). По умолчанию 1e-6.
    max_iter : int, optional
        Максимальное количество итераций для поиска минимума. По умолчанию 1000.

    Returns
    -------
    float or None
        Приближенное значение точки минимума, если метод сошелся.
        None, если достигнуто максимальное количество итераций.

    Examples
    --------
    >>> f = lambda x: x**2 - 4*x - 1
    >>> dichotomy_method(f, 0, 5)
    2.0000001788139343
    
    >>> f = lambda x: math.sin(x)
    >>> dichotomy_method(f, math.pi/2, 11*math.pi/6)
    4.712389479727394 # Апроксимация 3pi/2

    Notes
    -----
    1. Метод требует, чтобы функция f(x) была унимодальной на интервале [a, b] [[2]_].
    2. В текущей реализации присутствует проверка f(a)*f(b) ≥ 0, которая может быть необязательной для задач минимизации.
    3. Критерий остановки: длина интервала (b - a) < eps.
    4. Метод эффективен для гладких функций с одним минимумом, но может расходиться при нарушении унимодальности.

    References
    ----------
    .. [2] "Метод дихотомии - Википедия", https://en.wikipedia.org/wiki/Dichotomy
    .. [6] "Метод дихотомии (половинного деления)", https://example.com/dichotomy-method
    .. [7] "2.1 Метод дихотомии", https://example.com/dichotomy-theory
    """
    if f(a) * f(b) >= 0:
        raise ValueError("Функция не меняет знак на интервале [a, b]")

    for _ in range(max_iter):
        c = (a + b) / 2
        if (b - a) < eps:
            return c
        
        # Сравнение значений функции с обеих сторон от центра
        if f(c - eps/2) < f(c + eps/2):
            b = c
        else:
            a = c
            
    return None
####################################################################################
EQS = [bisection_method,bisection_method_params,dichotomy_method, newton_method,secant_method,wegstein_method,simple_iteration_method]


from ...forall import *

MTX = []

def mtx1(a, b, c):
    '''
    Задайте одномерный массив аг1 размерности, состоящий из {a} случайных целых чисел в пределах от {b} до {c}. Получите массив индексов, отсортированных по убыванию элементов массива. Выведите на печать массив ar1 с отсортированными элементами. Решить задачу средствами numpy и/или pandas. Не использовать циклы и конструкции стандартного Python там, где можно использовать возможности данных библиотек.
    Args: a,b,c
    '''
    import numpy as np


    ar1 = np.random.randint(a, b, c)


    sorted_indices = np.argsort(-ar1)


    sorted_ar1 = ar1[sorted_indices]


    print("Исходный массив:", ar1)
    print("Индексы по убыванию:", sorted_indices)
    print("Отсортированный массив:", sorted_ar1)

def mtx2(a,b,c,d):
    '''Создать матрицу {a} на {b} из случайных целых (используя модуль 'numpy.random) чисел из диапозона от {c} до {d} и найти в ней строку (ее индекс и вывести саму, строку), в которой сумма значений минимальна.'''
    matrix = np.random.randint(c, d, size=(a, b))


    row_sums = np.sum(matrix, axis=1)


    min_row_index = np.argmin(row_sums)

    min_row = matrix[min_row_index]


    print("Матрица 8x10:")
    print(matrix)
    print("Индекс:",min_row_index)
    print("Сама строка", min_row)


def mtx3():
    ''' Без непосредственного задания элементов матрицы создать в питру матрицу 11 на 6 вида: [[1, 2, 4, 7, 8, 9], [11, 12, 14, 17, 18, 19], (21, 22, 24, 27, 28, 29], ..., [101, 102, 104, 107, 108, 109]]'''
    start_values = np.arange(1, 102, 10).reshape(-1, 1)

    matrix = start_values + np.array([0, 1, 3, 6, 7, 8])

    print("Матрица 11×6:")
    print(matrix)

def mtx4():
    '''Решить матричное уравнение "A*X*B--С - найти матрицу Х. Где 'A - [[-1, 2, 4], [-3, 1, 2], 1-3, 0, 1]], В-[[3, -1], [2, 1]]`, `С-[[7, 21], [11, 8], [8, 4]].'''
    A = np.array([[-1, 2],
                [4, -3]])
    B = np.array([[1, 0],
                [0, 1]])
    C = np.array([[2, 3],
                [0, 1]])


    try:
        A_inv = np.linalg.inv(A)
        B_inv = np.linalg.inv(B)
    except np.linalg.LinAlgError:
        print("Одна из матриц A или B вырождена (не имеет обратной матрицы)")
    else:
        X = A_inv @ C @ B_inv
        print("Матрица X:")
        print(X)

def mtx5(a,b,c,d):
    '''Задать два двумерных массива аг1 и аг2 размерности ({a},{b}), состоящих из случайных целых чисел в пределах от {c} до {d}. Удвоить все значения аг1, которые больше значений ar2, расположенных на аналогичных позициях, остальные значения сделать равными
0. '''
    import numpy as np


    ar1 = np.random.randint(c, d+1, size=(a, b))
    ar2 = np.random.randint(c, d+1, size=(a, b))

    result = np.where(ar1 > ar2, ar1 * 2, 0)

    print("Массив ar1:\n", ar1)
    print("Массив ar2:\n", ar2)
    print("Результат:\n", result)

def mtx6(a,b,c,d):
    '''Задайте двумерный массив аг1 размерности ({a}, {b}), состоящий из случайных целых чисел в пределах от {b} до {c}. Определите, в каких столбцах не менее 2 раз встречается значение, максимальное по своей строке и распечатайте массив с заменой элементов остальных столбцов на 1. Решить задачу средствами пumpy и/или pandas. He использовать циклы и конструкции стандартного Python там, где можно использовать возможности данных библиотек.'''
    ar1 = np.random.randint(c, d+1, size=(a, b))
    print("Исходный массив:")
    print(ar1)


    max_in_rows = np.max(ar1, axis=1, keepdims=True)


    is_max = (ar1 == max_in_rows)


    max_counts = np.sum(is_max, axis=0)


    valid_columns = max_counts >= 2


    mask = np.zeros_like(ar1, dtype=bool)
    mask[:, valid_columns] = True


    result = np.where(mask, ar1, -1)

    print("\nРезультат:")
    print(result)


def mtx7(a,b,c,d):
    '''Задайте двумерный массив аг1 размерности, состоящий из случайных целых чисел в пределах от {a} до {b} размерности ({c},{d}). Получите массив, в котором нечетные элементы заменены на -1. Решить задачу средствами numpy и/или pandas. Не использовать циклы и конструкции стандартного Python там, где можно использовать возможности данных библиотек.'''
    ar1 = np.random.randint(a, b+1, size=(c, d))

    print("Исходный массив:")
    print(ar1)


    result = np.where(ar1 % 2 != 0, -1, ar1)

    print("\nМассив с заменой нечетных элементов на -1:")
    print(result)

def mtx8(a,b,c, d):
    '''Сгенерировать двухмерный массив `arr` размерности ({a}, {b}), состоящий из случайных действительных чисел, равномерно распределенных в диапазоне от {c} до {d}.
Нормализовать значения массива с помощью преобразования вида $ax+b$ так, что после нормализации максимальный элемент массива будет равен 1.0, минимальный 0.0'''
    import numpy as np

    arr = np.random.uniform(c, d, size = (a, b))
    print(arr, '\n')
    min_val = np.min(arr)
    max_val = np.max(arr)
    a = 1 / (max_val - min_val)
    b = -a * min_val
    norm_arr = a * arr + b
    print(norm_arr)

    print("\nМаксимальный элемент нормализованного массива:", np.max(norm_arr))
    print("Минимальный элемент нормализованного массива:", np.min(norm_arr))
    
MTX = [mtx1,mtx2,mtx3,mtx4,mtx5,mtx6,mtx7,mtx8]


from ...forall import *
#######################################################################################################################
# Условные характеристики относительно группы событий
#######################################################################################################################
def CCE_1(a,b):
    """Несимметричная игральная кость подбрасывается до тех пор, пока не выпадут цифры `V1` и `V2`.
    Пусть  X — число сделанных при этом бросков. Даны вероятности появления цифр в одном броске:
    P(`V1`)=`a` и P(`V2`)=`b`.Требуется найти:
    - E(X);
    - Var(X), если известно, что из `V1` и `V2` сначала выпала цифра `V2`

    Args:
        a (numerical): Вероятность выпадения 1-го числа
        b (numerical): Вероятность выпадения 2-го числа

    ## Prints
        `answer` каждое значение по очереди.<br>C запятой вместо точки и сокращением до соответствующего количества знаков после запятой

    Returns:
        `answer` (tuple): Соответствующие величины
    """
    ex = 1/a+1/b-1/(a+b)
    var=(1-a-b)/(a+b)**2+(1-a)/a**2
    
    answer = (ex, var)
    
    print('E(X)   = ' + rrstr(answer[0],4))
    print('Var(X) = ' + rrstr(answer[1],3))

    return answer
#######################################################################################################################
def CCE_2(a,b,c):
    """Несимметричная игральная кость подбрасывается до тех пор,
    пока не выпадут цифры `V1`, `V2` и `V3`.
    Пусть X — число сделанных при этом бросков.
    Даны вероятности появления цифр в одном броске: P(`V1`)=`a`, P(`V2`)=`b` и P(`V3`)=`c`. Требуется найти:
    - E(X);
    - Var(X), если известно, что из `V1`, `V2` и `V3` сначала выпала цифра `V1`, затем — `V2`

    Args:
        a (numerical): Вероятность выпадения 1-го числа
        b (numerical): Вероятность выпадения 2-го числа
        c (numerical): Вероятность выпадения 3-го числа
        
    ## Prints
        `answer` каждое значение по очереди.<br>C запятой вместо точки и сокращением до соответствующего количества знаков после запятой

    Returns:
        `answer` (tuple): Соответствующие величины
    """
    ex = (1+a/(b+c)+b/(a+c)+c/(a+b)+a*b/(c*(b+c))+a*b/(c*(a+c))+a*c/(b*(b+c))+a*c/(b*(b+a))+c*b/(a*(b+a))+c*b/(a*(c+a)))/(a+b+c)
    var =(1-a-b-c)/(a+b+c)**2+(1-b-c)/(b+c)**2+(1-c)/c**2 #при a->b->c
    
    answer = (ex, var)
    
    print('E(X)   = ' + rrstr(answer[0],4))
    print('Var(X) = ' + rrstr(answer[1],3))

    return answer
#######################################################################################################################
CCE = [CCE_1,CCE_2]


import re
from ..forall import *
from .fpb import *                      # Формулы полной вероятности и Байеса
from .sdrv import *                     # Специальные дискретные случайные величины
from .crv import *                      # Непрерывные случайные величины
from .nrv import *                      # Нормальные случайные векторы
from .anrv import *                     # Нормальные случайные векторы ДОПОЛНИТЕЛЬНЫЙ ПАКЕТ ФУНКЦИЙ
from .cce import *                      # Условные характеристики относительно группы событий
from .acmk import *                     # Приближенное вычисление вероятности методом Монте-Карло
from .pan import *                      # Портфельный анализ с невырожденной ковариационной матрицей
from .dt import *                       # Описательная статистика
from .ec import *                       # Эмперические характеристики
from .sffp import *                     # Выборки из конечной совокупности
from .mlm import *                      # Метод максимального правдоподобия
from .tmh import *                      # Проверка гипотез о значении среднего
from .tvh import *                      # Проверка гипотез о значении дисперсии
from .tsm import *                      # Проверка гипотезы о состоянии двух и трёх средних
from .cicc import *                     # Доверительный интервал для коэффициента корреляции
from .theory import *                   # Теоретические материалы

def printcolab():
    """Выводит ссылку на знакомый гугл колаб(google colab)"""
    print(r'https://colab.research.google.com/drive/1QjC3HnOivbi-38CvI7c9wq0mAYuXNQei?usp=sharing')
    return r'https://colab.research.google.com/drive/1QjC3HnOivbi-38CvI7c9wq0mAYuXNQei?usp=sharing'

def imports():
    """
    Возвращает строку, содержащую инструкции импорта Python для различных научных библиотек. 
    Эти библиотеки обычно используются для математических вычислений, символьной математики, 
    и комбинаторные операции с инициализацией SymPy для красивой печати с использованием Unicode. 
    и форматирование LaTeX.
    """
    print('''
    
from scipy.integrate import quad
import math
import numpy as np
import sympy
import itertools
sympy.init_printing(use_unicode=True,use_latex=True)
    ''')
    return '''
    
from scipy.integrate import quad
import math
import numpy as np
import sympy
import itertools
sympy.init_printing(use_unicode=True,use_latex=True)
    '''

UF = [printcolab,imports]

files_dict ={
    'Полезные функции': UF,
    'Теоретические материалы':THEORY,
    'Формулы полной вероятности и Байеса':FPB,
    'Специальные дискретные случайные величины':SDRV,
    'Непрерывные случайные величины':CRV,
    'Нормальные случайные векторы':NRV,
    'Нормальные случайные векторы ДОПОЛНИТЕЛЬНЫЙ ПАКЕТ ФУНКЦИЙ':ANRV,
    'Условные характеристики относительно группы событий':CCE,
    'Приближенное вычисление вероятности методом Монте-Карло':ACMK,
    'Портфельный анализ с невырожденной ковариационной матрицей':PAN,
    'Описательная статистика':DT,
    'Эмперические характеристики':EC,
    'Выборки из конечной совокупности':SFFP,
    'Метод максимального правдоподобия':MLM,
    'Проверка гипотез о значении среднего':TMH,
    'Проверка гипотез о значении дисперсии': TVH,
    'Проверка гипотезы о состоянии двух и трёх средних': TSM,
    'Доверительный интервал для коэффициента корреляции':CICC
}

names = list(files_dict.keys())
modules = list(files_dict.values())


    
def enable_ppc():
    """
    Returns a string containing a Python script that uses the pyperclip module to
    define a function named `write`. The `write` function takes a single argument `name`,
    copies it to the system clipboard, and pastes it using pyperclip.
    """
    return'''
import pyperclip

#Делаем функцию которая принимает переменную text
def write(name):
    pyperclip.copy(name) #Копирует в буфер обмена информацию
    pyperclip.paste()'''
    
def invert_dict(d):
    """
    Returns a new dictionary with the keys and values of the input dictionary swapped.
    
    Example:
        >>> invert_dict({1: 'a', 2: 'b'})
        {'a': 1, 'b': 2}
    """
    return {value: key for key, value in d.items()}

# Создание словарей функций для каждой темы
# funcs_dicts: {описание_задачи: функция}
funcs_dicts = [dict([(get_task_from_func(i), i) for i in module]) for module in modules]
# funcs_dicts_ts: {описание_задачи_без_пробелов_и_переносов: функция} (ts - to_search)
funcs_dicts_ts = [dict([(get_task_from_func(i,True), i) for i in module]) for module in modules]
# funcs_dicts_full: {имя_функции: исходный_код_функции}
funcs_dicts_full = [dict([(i.__name__, getsource(i)) for i in module]) for module in modules]
# funcs_dicts_full_nd: {имя_функции: исходный_код_функции_без_документации}
funcs_dicts_full_nd = [dict([(i.__name__, getsource_no_docstring(i)) for i in module]) for module in modules]

# Создание словарей, группирующих функции по темам
themes_list_funcs = dict([(names[i],list(funcs_dicts[i].values()) ) for i in range(len(names))])        # Название темы : список функций по теме
themes_list_dicts = dict([(names[i],funcs_dicts[i]) for i in range(len(names))])                        # Название темы : словарь по теме, где ЗАДАНИЕ: ФУНКЦИИ
themes_list_dicts_full = dict([(names[i],funcs_dicts_full[i]) for i in range(len(names))])              # Название темы : словарь по теме, где НАЗВАНИЕ ФУНКЦИИ: ТЕКСТ ФУНКЦИИ
themes_list_dicts_full_nd = dict([(names[i],funcs_dicts_full_nd[i]) for i in range(len(names))])        # Название темы : словарь по теме, где НАЗВАНИЕ ФУНКЦИИ: ТЕКСТ ФУНКЦИИ БЕЗ ДОКУМЕНТАЦИИ


# Тема -> Функция -> Задание
def description(
    dict_to_show=themes_list_funcs,
    key=None,
    show_only_keys: bool = False,
    show_keys_second_level: bool = True,
    n_symbols: int = 32,
    to_print: bool = True,
    show_doc = False):
    """
    Отображает информацию о доступных функциях и темах в модуле TViMS.

    Функция предоставляет несколько режимов для просмотра содержимого модуля:
    - Показать все темы и функции в них.
    - Показать функции в рамках одной темы.
    - Показать исходный код конкретной функции.

    Args:
        dict_to_show (str or dict, optional): Словарь для отображения или название темы.
            По умолчанию `themes_list_funcs`.
        key (hashable, optional): Ключ для фильтрации (например, имя функции).
        show_only_keys (bool, optional): Если True, показывать только ключи (названия тем).
            По умолчанию False.
        show_keys_second_level (bool, optional): Если True, показывать ключи второго уровня (названия функций).
            По умолчанию True.
        n_symbols (int, optional): Максимальное количество символов для описания.
            По умолчанию 32.
        to_print (bool, optional): Если True, выводит результат в консоль. Иначе возвращает строку.
            По умолчанию True.
        show_doc (bool, optional): Если True, показывает полный исходный код функции с документацией.
            По умолчанию False.

    Returns:
        str or None: Форматированная строка с информацией, если `to_print` равно False.
            В противном случае None.

    Examples:
        >>> # Показать только названия тем в модуле TViMS
        >>> description(show_only_keys=True)

        >>> # Показать функции и их краткое описание для темы 'Описательная статистика'
        >>> description('Описательная статистика')

        >>> # Показать исходный код функции 'describe_text'
        >>> description('Описательная статистика', key='describe_text', show_doc=True)

    Notes:
        - Функция использует предопределенные словари `themes_list_funcs`, `themes_list_dicts` и др.
          для доступа к метаданным функций.
        - Для корректного отображения длинных описаний используется автоматический перенос строк.
    """
    
    # Если dict_to_show - строка (название темы) и не указан конкретный ключ (key)
    if type(dict_to_show) == str and key == None:
        dict_to_show = themes_list_dicts[dict_to_show]
        dict_to_show = invert_dict(dict_to_show)
        text = ""
        length1 = 1 + max([len(x.__name__) for x in list(dict_to_show.keys())])
        
        for key in dict_to_show.keys():
            text += f'{key.__name__:<{length1}}' # Имя функции, выровненное по левому краю
            
            if not show_only_keys:
                text += ': '
                text += f'{dict_to_show[key]};\n' + ' '*(length1+2) # Описание задачи
            text += '\n'
            
        if to_print == True:
            return print(text)
        return text
    
    # Если dict_to_show - строка (название темы) и указан конкретный ключ (имя функции)
    elif type(dict_to_show) == str and key in themes_list_dicts_full[dict_to_show].keys():
        if show_doc:
            return print(themes_list_dicts_full[dict_to_show][key]) # Вывод исходного кода функции
        else:
            return print(themes_list_dicts_full_nd[dict_to_show][key]) # Вывод исходного кода функции
    
    else:
        show_only_keys = False
    text = ""
    length1 = 1 + max([len(x) for x in list(dict_to_show.keys())]) # Максимальная длина ключа первого уровня (названия темы)
    
    for key in dict_to_show.keys():
        text += f'{key:^{length1}}' # Название темы, выровненное по центру
        if not show_only_keys:
            text += ': '
            for f in dict_to_show[key]:
                text += f'{f.__name__}'
                if show_keys_second_level:
                    text += ': '
                    try:
                        # Получение описания функции из инвертированного словаря
                        func_text_len = len(invert_dict(themes_list_dicts[key])[f])
                        
                        # Форматирование описания с переносами строк и ограничением по длине
                        func_text = invert_dict(themes_list_dicts[key])[f]
                        text += func_text.replace('\n','\n'+' '*(length1 + len(f.__name__))) if func_text_len < n_symbols else func_text[:n_symbols].replace('\n','\n'+' '*(length1 + len(f.__name__)))+'...'
                    except:
                        pass # Пропуск, если описание не найдено
                text += ';\n' + ' '*(length1+2) # + '\n' + ' '*(length1+2)
        text += '\n'
        
    if to_print == True:
        return print(text)
    return text



def search(query: str, to_print: bool = True, data: str = description(n_symbols=10000, to_print=False)):
    """
    Ищет функции в модуле TViMS по ключевым словам в их описаниях.

    Функция выполняет регистронезависимый поиск по описаниям всех доступных
    элементов и выводит найденные совпадения.

    Args:
        query (str): Строка для поиска. Может быть частью слова или фразой.
        to_print (bool, optional): Если True, результат выводится в консоль.
            Иначе возвращается в виде списка строк. По умолчанию True.
        data (str, optional): Строка с данными для поиска, обычно результат
            вызова `description(to_print=False)`. По умолчанию генерируется
            автоматически.

    Returns:
        list or None: Список найденных строк в формате "Тема : Описание",
            если `to_print` равно False. В противном случае None.

    Examples:
        >>> # Найти все, что связано с гипотезами
        >>> search("гипотез")

        >>> # Найти функции для работы с распределениями и вернуть результат в виде списка
        >>> results = search("распределен", to_print=False)
        >>> print(results)

    Notes:
        - Поиск чувствителен к структуре данных, предоставляемых функцией `description`.
        - Для полного поиска по всем описаниям `description` вызывается с `n_symbols=10000`.
    """
    # Разделение входных данных на отдельные темы
    topics = re.split(r'\n\s*\n', data)
    matches = []

    for topic_data in topics:
        # Пропуск пустых блоков тем
        if not topic_data.strip():
            continue

        topic_match = re.match(r'^\s*(.*?):', topic_data)
        if not topic_match:
            continue
        
        # Извлечение названия темы
        topic = topic_match.group(1).strip()
        # Поиск всех функций и их описаний в текущей теме
        functions = re.findall(r'(\w+)\s*:\s*([\s\S]*?)(?=\n\s*\w+\s*:|\Z)', topic_data)

        for func, description in functions:
            # Проверка наличия запроса (без учета регистра) в описании функции
            if query.lower() in description.lower():
                matches.append(f"{topic} : {description.strip()}")
    
    # Вывод результатов или их возврат списком
    if to_print:
        return print("\n".join(matches))
    return matches


import numpy as np
import pandas as pd
import re
from inspect import getsource, getdoc
import inspect
import ast
import matplotlib.pyplot as plt
import sys
from numba import njit, prange
import warnings
warnings.filterwarnings('ignore')
####################################################################################
pattern = r'\s*(.*?)\s*(?=def __init__|Args|Параметры|Parameters)' #     pattern = r'\s*[\s\S]*?([\s\S]+?)\s+[Parameters|Args]'
####################################################################################
def get_task_from_func(func, to_search=False):
    """
    Извлекает описание параметров или полный docstring из функции.

    Parameters
    ----------
    func : callable
        Функция, из которой извлекается документация.
    to_search : bool, optional
        Если True, ищет только секцию 'Args' или 'Parameters' и удаляет пробелы/переносы.
        Если False, возвращает полный docstring с сохранением форматирования.

    Returns
    -------
    str or None
        Обработанная часть docstring или None, если документация отсутствует.

    Notes
    -----
    1. Использует регулярные выражения для поиска секции 'Args'/'Parameters'.
    2. При отсутствии docstring возвращает None, чтобы избежать ошибок.
    3. Рекомендуется использовать `inspect.getdoc()` вместо регулярного выражения для полного текста.

    Examples
    --------
    >>> def example(x: int):
    ...     '''Example function.
    ...     Parameters
    ...     ----------
    ...     x : int
    ...         Value to process.
    ...     '''
    ...     return x
    ...
    >>> get_task_from_func(example, to_search=True)
    'Parameters----------x:intValueToprocess.'

    >>> get_task_from_func(example, to_search=False)  # doctest: +ELLIPSIS
    'Example function.\\n    Parameters\\n    ----------\\n    x : int\\n        Value to process.'
    """
    doc = getdoc(func)
    if not doc:
        return None
    
    try:
        return re.search(pattern,doc,re.DOTALL).group(1).replace('\n','').replace(' ','') if to_search else re.search(pattern,doc,re.DOTALL).group(1)
    except:
        try:
            return doc.replace('\n','').replace(' ','') if to_search else doc
        except:
            return doc
####################################################################################
#from .nm.additional_funcs import *
####################################################################################
# rrstr (Округление до n знаков)
####################################################################################
def one_rrstr(x, n=0): # округление до n знаков после запятой
    if n == 0:
        return str(x)
    fmt = '{:.' + str(n) + 'f}'
    return fmt.format(x).replace('.', ',')

def un_one_rrstr(x):
    return float(x.replace(',', '.'))

def un_rrstr(x):
    return np.vectorize(un_one_rrstr)(x)


def rrstr(x, n):
    """
    Форматирует числа или массивы чисел с округлением до заданного количества знаков после запятой.

    Parameters
    ----------
    x : float or array_like
        Число или массив чисел для форматирования.
    n : int
        Количество знаков после десятичного разделителя (запятой).

    Returns
    -------
    str or ndarray
        Строка (для скалярных входов) или массив строк (для последовательностей).
        Десятичный разделитель представлен запятой.

    Notes
    -----
    1. Поддерживает векторизацию через numpy.vectorize.
    2. Для n=0 выполняет только замену точки на запятую (если есть).
    3. Требует установленной библиотеки numpy.
    4. Не предназначен для работы с отрицательными значениями n.

    Examples
    --------
    >>> rrstr(3.1415, 2)
    '3,14'

    >>> rrstr([1.2345, 2.789], 1)
    array(['1,2', '2,8'], dtype='<U3')

    >>> rrstr(42, 0)
    '42'

    References
    ----------
    .. [1] NumPy Documentation: https://numpy.org/doc/
    .. [2] Python Software Foundation. "Python Language Reference", version 3.11.
    """
    rrstr1 = np.vectorize(one_rrstr)
    res = rrstr1(x, n)
    if res.size == 1:
        return str(res)
    return res
####################################################################################
# show_img показывает фотографию в ячейке вывода
####################################################################################
def show_img(filename):
    """
    Отображает изображение в интерфейсе Jupyter Notebook/IPython.

    Parameters
    ----------
    filename : str
        Путь к файлу изображения, который необходимо отобразить.
        Поддерживаются стандартные графические форматы (PNG, JPEG и др.).

    Returns
    -------
    str
        Возвращает переданный путь к файлу без изменений, 
        независимо от успешности отображения изображения.

    Notes
    -----
    1. Требует наличия библиотеки IPython в окружении.
    2. Функция работает только в интерактивных средах (Jupyter Notebook, IPython).
    3. При отсутствии файла/ошибке загрузки:
       - Выводит сообщение 'Неправильное имя файла'
       - Не прерывает выполнение программы
    4. Поддерживает все форматы, поддерживаемые классом IPython.display.Image

    Examples
    --------
    >>> show_img("example.png")
    # Отображает изображение example.png если оно существует

    >>> show_img("nonexistent.jpg")
    Неправильное имя файла
    'nonexistent.jpg'

    References
    ----------
    .. [1] IPython Documentation: https://ipython.readthedocs.io/en/stable/
    .. [2] Jupyter Notebook documentation: https://jupyter-notebook.readthedocs.io/en/latest/
    """
    from IPython.display import display, Image
    try:
        img = Image(filename=filename)
        display(img)
    except:
        print('Неправильное имя файла')
    return filename
####################################################################################
# show_images показывает несколько фотографий 
# в ячейке вывода по названиям в итерируемом аргументе  
####################################################################################
def show_images(filenames):
    """
    Отображает изображения из указанных файлов в интерфейсе Jupyter Notebook/IPython.

    Параметры
    ---------
    filenames : str or array_like
        Путь к одному файлу изображения (строка) или последовательность путей (список, массив).
        Поддерживаются стандартные графические форматы (PNG, JPEG и др.).

    Возвращает
    ----------
    str or ndarray
        Если входной аргумент - скаляр: возвращает строку с именем файла.
        Если входной аргумент - последовательность: возвращает numpy.ndarray со списком имен файлов.

    Примечания
    ---------
    1. Функция основана на `np.vectorize` для поддержки массивов/списков файлов.
    2. Требует наличия библиотеки IPython и работы в интерактивной среде (Jupyter Notebook, IPython).
    3. При отсутствии файла/ошибке загрузки:
       - Выводит сообщение 'Неправильное имя файла'
       - Не прерывает выполнение программы
       - Возвращает имя файла как есть
    4. Каждое изображение отображается немедленно при обработке.

    Примеры
    --------
    >>> show_images("example.png")
    'example.png'  # Изображение будет отображено в ячейке вывода

    >>> show_images(["img1.jpg", "img2.png"])
    array(['img1.jpg', 'img2.png'], dtype='<U8')  # Оба изображения отобразятся последовательно

    >>> show_images(np.array([["a.png", "b.png"], ["c.png", "d.png"]]))
    array([['a.png', 'b.png'], ['c.png', 'd.png']], dtype='<U5')  # Поддержка многомерных массивов

    Ссылки
    ------
    .. [1] Документация show_img: https://github.com/example/docs/show_img
    .. [2] IPython Documentation: https://ipython.readthedocs.io/en/stable/
    .. [3] Jupyter Notebook documentation: https://jupyter-notebook.readthedocs.io/en/latest/
    """
    return np.vectorize(show_img)(filenames)
####################################################################################
# save_pdf_as_images (Сохраняет каждую страничку pdf как png файл)
####################################################################################    
def save_pdf_as_images(pdf_path, output_folder=None, dpi=100):
    """
    Конвертирует PDF-документ в изображения и сохраняет их в указанной папке.

    Parameters
    ----------
    pdf_path : str
        Путь к исходному PDF-файлу, который необходимо конвертировать.
    output_folder : str, optional
        Путь к папке, куда будут сохранены изображения. Если None, создается папка
        с тем же именем, что и у PDF-файла, но без расширения.
    dpi : int, optional
        Разрешение вывода изображений в точках на дюйм (dots per inch).
        Более высокое значение увеличивает качество, но замедляет обработку.

    Returns
    -------
    list of str
        Список абсолютных путей к сохраненным PNG-файлам.

    Notes
    -----
    1. Требует установленной библиотеки `pdf2image` и Poppler (для Windows).
    2. Автоматически создает выходную папку, если она отсутствует.
    3. Имена файлов генерируются по шаблону "page_номер.png".
    4. При работе с большими PDF файлами потребуется больше оперативной памяти.
    5. Для Linux-систем Poppler можно установить через пакетный менеджер.

    Examples
    --------
    >>> save_pdf_as_images("document.pdf")
    # Сохранит страницы в папке "document" с разрешением 100 DPI
    
    >>> save_pdf_as_images("report.pdf", "output_images", dpi=200)
    # Сохранит страницы в папке "output_images" с разрешением 200 DPI

    References
    ----------
    .. [1] "pdf2image documentation" https://pdf2image.readthedocs.io/en/latest/
    .. [2] Python Software Foundation. "Python Language Reference", version 3.11.
    """
    from pdf2image import convert_from_path
    import os
    
    if output_folder is None:
        output_folder = pdf_path[:-4]
        
    outpaths = []
    # Создаем папку для сохранения, если её нет
    os.makedirs(output_folder, exist_ok=True)

    # Конвертируем PDF в список изображений
    try:
        pages = convert_from_path(pdf_path, dpi=dpi)
    except Exception as e:
        raise RuntimeError(f"Ошибка при конвертации PDF: {str(e)}")

    # Сохраняем каждую страницу как изображение
    for page_number, page in enumerate(pages, start=1):
        output_path = os.path.join(output_folder, f"page_{page_number}.png")
        try:
            page.save(output_path, "PNG")
            outpaths.append(os.path.abspath(output_path))
            print(f"Страница {page_number} сохранена как {output_path}")
        except Exception as e:
            print(f"Ошибка при сохранении страницы {page_number}: {str(e)}")
            
    return outpaths
####################################################################################
import pyperclip

#Делаем функцию которая принимает переменную text
def write(name):
    """
    Копирует указанную информацию в системный буфер обмена.

    Parameters
    ----------
    name : str
        Строка, которую необходимо скопировать в буфер обмена.

    Returns
    -------
    None
        Функция не возвращает значение, но выполняет побочный эффект (копирование в буфер).

    Notes
    -----
    1. Функция использует библиотеку `pyperclip`, которая должна быть установлена.
    2. Второй вызов `pyperclip.paste()` не влияет на результат выполнения функции.
    3. Для работы на Linux-системах может потребоваться установка дополнительных зависимостей (например, `xclip`).

    Examples
    --------
    >>> write("Hello, World!")
    # Строка "Hello, World!" будет скопирована в буфер обмена

    >>> write("12345")
    # Числовая строка "12345" будет доступна для вставки из буфера

    References
    ----------
    .. [1] Pyperclip documentation: https://pyperclip.readthedocs.io/en/latest/
    .. [2] Sweigart, A. "Automate the Boring Stuff with Python", 2nd edition.
    """
    pyperclip.copy(name)  # Копирование в буфер обмена
    pyperclip.paste()     # Дублирующий вызов для проверки буфера
####################################################################################
def invert_dict(d):
    """
    Возвращает новый словарь с поменяными местами ключами и значениями исходного словаря.

    Parameters
    ----------
    d : dict
        Исходный словарь, значения которого должны быть хэшируемыми (например, не списки).

    Returns
    -------
    dict
        Словарь, где ключи и значения исходного словаря поменяны местами.

    Notes
    -----
    1. Если в исходном словаре есть дублирующиеся значения, последние ключи будут перезаписаны.
    2. Функция не поддерживает словари с нехэшируемыми значениями (например, списками или множествами).
    3. Используется для упрощения двустороннего поиска в словарях с уникальными значениями.

    Examples
    --------
    >>> invert_dict({'a': 1, 'b': 2})
    {1: 'a', 2: 'b'}

    >>> invert_dict({'x': 'val', 'y': 'val'})
    {'val': 'y'}  # Второй ключ перезаписывает первый

    References
    ----------
    .. [1] Python Software Foundation. "Python Language Reference", version 3.11.
    .. [2] Beazley, D.M. "Python Essential Reference", 4th edition.
    """
    return {value: key for key, value in d.items()}
####################################################################################
def getsource_no_docstring(obj):
    """
    Возвращает исходный код объекта без его строки документации (docstring).
    Аналогична inspect.getsource, но удаляет docstring.

    Выбрасывает те же исключения, что и inspect.getsourcelines/inspect.getsource,
    если исходный код не может быть получен.
    """
    try:
        # sourcelines - это список строк с оригинальными окончаниями
        # lnum - номер начальной строки в исходном файле (здесь не используется)
        sourcelines, lnum = inspect.getsourcelines(obj)
    except (TypeError, OSError) as e:
        # Если inspect.getsourcelines не смогли получить код
        raise e # Перевыбрасываем оригинальное исключение

    source_code_str = "".join(sourcelines)

    try:
        # Парсим весь блок кода, полученный для объекта
        tree = ast.parse(source_code_str)
    except SyntaxError:
        # Если код объекта по какой-то причине не является валидным Python
        # (маловероятно для вывода getsourcelines, но для полноты)
        # возвращаем исходный код "как есть".
        return source_code_str

    docstring_expr_node = None
    
    # Узел, содержащий тело, где может быть docstring (модуль, функция, класс)
    node_with_body = None

    if inspect.ismodule(obj):
        # Если объект - модуль, то tree это ast.Module. Его тело - tree.body.
        node_with_body = tree
    elif tree.body and isinstance(tree.body[0], (ast.FunctionDef, ast.ClassDef, ast.AsyncFunctionDef)):
        # Если объект - функция или класс, то tree.body[0] - это узел FunctionDef/ClassDef.
        # Его тело - tree.body[0].body.
        node_with_body = tree.body[0]
    else:
        # Неожиданная структура AST или getsource вернул что-то не то.
        return source_code_str

    if node_with_body and hasattr(node_with_body, 'body') and node_with_body.body:
        first_statement_in_body = node_with_body.body[0]
        if isinstance(first_statement_in_body, ast.Expr):
            value_node = first_statement_in_body.value
            # Проверяем, является ли значение строковым литералом
            is_string_literal = False
            if isinstance(value_node, ast.Str): # Для Python < 3.8
                is_string_literal = True
            elif hasattr(ast, 'Constant') and isinstance(value_node, ast.Constant): # Для Python 3.8+
                if isinstance(value_node.value, str):
                    is_string_literal = True
            
            if is_string_literal:
                docstring_expr_node = first_statement_in_body

    if docstring_expr_node:
        # Номера строк в AST (lineno, end_lineno) 1-базированные и относятся к началу
        # строки source_code_str (которая соответствует sourcelines).
        doc_start_line_idx = docstring_expr_node.lineno - 1  # 0-базированный индекс начала
        doc_end_line_idx = docstring_expr_node.end_lineno - 1    # 0-базированный индекс конца

        # Собираем новый список строк, исключая строки с docstring
        new_sourcelines = []
        for i, line_content in enumerate(sourcelines):
            if not (doc_start_line_idx <= i <= doc_end_line_idx):
                new_sourcelines.append(line_content)
        
        return "".join(new_sourcelines)
    else:
        # Docstring не найден в ожидаемом месте или отсутствует.
        return source_code_str
####################################################################################
from .ml.additional_funcs import *


from ...forall import *
def perepolnenie():
    """
    Демонстрация поведения типов данных при переполнении в Python и NumPy.

    Parameters
    ----------
    Нет входных параметров

    Returns
    ----------
    None
        Функция только выводит информацию в консоль.

    Examples
    --------
    >>> perepolnenie()
    2147483647
    -2147483648
    1.7976931348623157e+308
    0.0
    inf

    Notes
    -----
    1. Для int32 (32-битное целое):
    - Максимальное значение: 2^31-1 = 2147483647
    - При переполнении происходит "wrap-around" (переполнение с отрицательным результатом)
    
    2. Для чисел с плавающей точкой (float64):
    - Максимальное значение: ~1.8e308
    - Добавление к max_float приводит к:
        - 0.0 при малых приращениях (потеря точности)
        - inf при двойном max_float (положительная бесконечность)

    3. Отличия от целочисленной арифметики Python:
    - Стандартные int Python не имеют жестких ограничений
    - NumPy использует фиксированное количество бит для хранения

    References
    ----------
    .. [1] Официальная документация Python: https://docs.python.org/3/library/sys.html
    .. [2] NumPy reference manual: https://numpy.org/doc/stable/reference/
    .. [3] IEEE Standard for Floating-Point Arithmetic (IEEE 754)
    .. [4] Голуб Дж., Ван Лоун Ч. Матричные вычисления. М.: Мир, 1999.
    """
    print(f'переполнение')
    print(np.iinfo(np.int32).max)
    print(np.int32(np.iinfo(np.int32).max) + 1)
    print(sys.float_info.max)
    print(sys.float_info.max + 5 - sys.float_info.max)
    print(sys.float_info.max + sys.float_info.max)


def poterya_tochnosti():
    """
    Демонстрация потери точности при арифметических операциях с числами с плавающей точкой.

    Parameters
    ----------
    Нет входных параметров

    Returns
    -------
    None
        Функция только выводит информацию в консоль.

    Examples
    --------
    >>> poterya_tochnosti()
    потеря точности
    1e-09

    (В реальности возможен вывод 0.0 или 1e-09 в зависимости от реализации float в интерпретаторе)

    Notes
    -----
    1. Проблема возникает из-за ограниченной точности представления чисел с плавающей точкой (IEEE 754):
    - 64-битные числа (float64) имеют ~15-17 десятичных знаков точности
    - При вычитании близких чисел происходит потеря значащих разрядов (catastrophic cancellation)
    
    2. В примере:
    - 1.000_000_001 и 1.000_000_000 различаются на 1e-9
    - При представлении в формате float64 эти числа могут храниться с одинаковым приближением

    3. Для сравнения:
    - Целочисленная арифметика в Python не имеет этого ограничения
    - Для высокой точности можно использовать модуль decimal.Decimal

    References
    ----------
    .. [1] IEEE Standard for Floating-Point Arithmetic (IEEE 754)
    .. [2] Голуб Дж., Ван Лоун Ч. Матричные вычисления. М.: Мир, 1999.
    .. [3] Официальная документация Python по числам с плавающей точкой: https://docs.python.org/3/tutorial/floatingpoint.html
    .. [4] Higham, N. J. Accuracy and Stability of Numerical Algorithms. SIAM, 2002.
    """
    print(f'потеря точности')
    print(1.000_000_001 - 1.000_000_000)


def oshibka_okrugleniya():
    """
    Демонстрация ошибок округления при арифметических операциях с числами с плавающей точкой.

    Parameters
    ----------
    Нет входных параметров

    Returns
    -------
    None
        Функция только выводит информацию в консоль.

    Examples
    --------
    >>> oshibka_okrugleniya()
    ошибки округления
    False

    (Результат может варьироваться в зависимости от реализации float в интерпретаторе)

    Notes
    -----
    1. Проблема вызвана ограничениями стандарта IEEE 754 для чисел с плавающей точкой:
    - Числа 0.1 и 0.2 не могут быть точно представлены в двоичной системе
    - При сложении возникает накопление ошибки округления
    - Результат 0.1 + 0.2 ≈ 0.30000000000000004 (не равно 0.3)

    2. Подобные ошибки характерны для:
    - Все языков программирования, использующих IEEE 754
    - Операций с десятичными дробями, которые не являются степенью двойки
    - Сравнений на равенство с float числами

    3. Для точных вычислений:
    - Используйте модуль decimal.Decimal для финансовых расчетов
    - Применяйте сравнение с допуском (epsilon): abs(a - b) < 1e-9

    References
    ----------
    .. [1] IEEE Standard for Floating-Point Arithmetic (IEEE 754)
    .. [2] Голуб Дж., Ван Лоун Ч. Матричные вычисления. М.: Мир, 1999.
    .. [3] Официальная документация Python по числам с плавающей точкой: https://docs.python.org/3/tutorial/floatingpoint.html
    .. [4] Higham, N. J. Accuracy and Stability of Numerical Algorithms. SIAM, 2002.
    """
    print(f'ошибки округления')
    print(0.1 + 0.2 == 0.3)

    

def nakoplenie_oshibok():
    """
    Демонстрация накопления ошибок округления при многократном сложении чисел с плавающей точкой.

    Parameters
    ----------
    Нет входных параметров

    Returns
    -------
    None
        Функция только выводит информацию в консоль.

    Examples
    --------
    >>> nakoplenie_oshibok()
    накопление ошибок
    100000.0
    0.0

    (На самом деле возможен вывод вроде:
    100000.00000143294
    -1.43294e-06
    или другие незначительные отклонения от теоретического значения)

    Notes
    -----
    1. Проблема вызвана ограничениями стандарта IEEE 754:
    - Число 0.1 не может быть точно представлено в двоичной системе
    - Каждое сложение добавляет малую ошибку округления
    - При миллионе итераций эти ошибки кумулируются в значимую величину

    2. Особенности:
    - Даже при теоретическом результате 100_000.0 (1_000_000 * 0.1) 
        фактический результат может отличаться
    - Ошибка зависит от порядка операций и реализации float в интерпретаторе

    3. Для сравнения:
    - При использовании целочисленной арифметики (например, операций с копейками) 
        такой проблемы нет
    - Для финансовых расчетов рекомендуется использовать модуль decimal.Decimal
    - Для научных вычислений применяют алгоритмы компенсированного суммирования (алгоритм Кэхэна)

    References
    ----------
    .. [1] IEEE Standard for Floating-Point Arithmetic (IEEE 754)
    .. [2] Голуб Дж., Ван Лоун Ч. Матричные вычисления. М.: Мир, 1999.
    .. [3] Официальная документация Python по числам с плавающей точкой: https://docs.python.org/3/tutorial/floatingpoint.html
    .. [4] Higham, N. J. Accuracy and Stability of Numerical Algorithms. SIAM, 2002.
    """
    print(f'накопление ошибок')
    x = 0
    true_init = 100_000.0
    for _ in range(1_000_000):
      x+=0.1
    print(x)
    print(true_init-x)




def poterya_znachimosti():
    """
    Демонстрация потери значимости при вычислении экспоненты малых чисел.

    Parameters
    ----------
    Нет входных параметров

    Returns
    -------
    None
        Функция только выводит информацию в консоль.

    Examples
    --------
    >>> poterya_znachimosti()
    потеря значимости
    1.0000000827.40371e-10

    (Фактический результат может варьироваться в зависимости от точности вычислений NumPy)

    Notes
    -----
    1. Проблема вызвана комбинацией:
    - Представление чисел с плавающей точкой в формате IEEE 754 (ограниченная точность)
    - Вычитание близких значений (1 и exp(1e-10) ≈ 1.0000000001)

    2. Для малых x известно, что:
    exp(x) ≈ 1 + x + x²/2! + x³/3! + ...
    При x=1e-10, теоретическое значение exp(x)-1 ≈ 1e-10

    3. Ошибки проявляются как:
    - Потеря относительной точности из-за вычитания
    - Округление промежуточных результатов
    - Ограниченное количество значащих десятичных разрядов (15-17 для float64)

    4. Альтернативные подходы:
    - Использование специальной функции `np.expm1(x)` для точного вычисления exp(x)-1
    - Разложение в ряд Тейлора при малых x
    - Использование библиотек с произвольной точностью

    References
    ----------
    .. [1] IEEE Standard for Floating-Point Arithmetic (IEEE 754)
    .. [2] Голуб Дж., Ван Лоун Ч. Матричные вычисления. М.: Мир, 1999.
    .. [3] Официальная документация NumPy: https://numpy.org/doc/
    .. [4] Higham, N. J. Accuracy and Stability of Numerical Algorithms. SIAM, 2002.
    """
    print(f'потеря значимости')
    print(np.exp(1e-10) - 1)





def summirovanie_po_Kohanu(numbers):
    """
    Алгоритм суммирования по Кохану для минимизации ошибок округления при сложении чисел с плавающей точкой.
    Parameters
    ----------
    numbers : list or ndarray
        Последовательность чисел с плавающей точкой или целых чисел, которые необходимо просуммировать. 
        Пример: [1.0, 2.0, 3.0]
    Returns
    -------
    float
        Сумма всех чисел в последовательности с учетом компенсации ошибок округления по алгоритму Кохана.
    Examples
    --------
    >>> summirovanie_po_Kohanu([1.0, 2.0, 3.0])
    6.0
    >>> summirovanie_po_Kohanu(np.arange(1, 10**5))
    4999950000.0
    Notes
    -----
    Алгоритм Кохана позволяет сохранить численную точность при суммировании большого количества чисел с плавающей точкой, 
    что особенно важно при операциях с большими объемами данных или в научных вычислениях. 
    Метод использует дополнительную переменную для хранения накопленной ошибки округления и корректирует итоговую сумму на её значение.
    References
    ----------
    .. [1] Кахан В. (1965). "Further remarks on reducing truncation errors." Numerische Mathematik 7 (1): 488.
    .. [2] Higham, N. J. (2002). Accuracy and Stability of Numerical Algorithms (2nd ed.). SIAM.
    .. [3] Wikipedia: https://ru.wikipedia.org/wiki/Суммирование_по_Кохану
    """
    # Алгоритм суммирования по Кохану для минимизации ошибок округления
    def kahan_sum(numbers):
        total = 0.0
        correction = 0.0
        for num in numbers:
            y = num - correction
            temp = total + y
            correction = (temp - total) - y
            total = temp
        return total
    return kahan_sum(numbers)

BASE = [perepolnenie,poterya_tochnosti,oshibka_okrugleniya,nakoplenie_oshibok,poterya_znachimosti,summirovanie_po_Kohanu]


import pandas as pd
import numpy as np
import scipy.stats as st
from collections import defaultdict

class NormalNaiveBayes:
    """
    Наивный байесовский классификатор с предположением о нормальном (гауссовском) распределении признаков.
    
    Модель предполагает, что признаки для каждого класса распределены по нормальному закону,
    и оценивает параметры (среднее, стандартное отклонение) этих распределений по обучающей выборке.
    """
    
    def fit(self, X, y):
        """
        Обучает модель на основе обучающих данных.
        
        Args:
            X (array-like): Матрица признаков обучающей выборки.
            y (array-like): Вектор меток классов.
            
        Returns:
            NormalNaiveBayes: Обученный экземпляр классификатора.
        """
        # Объединяем признаки и метки в один набор для удобства группировки
        dataset = np.column_stack((X, y))
        
        # Группируем строки по метке класса с помощью defaultdict
        data_per_class = defaultdict(list)
        for row in dataset:
            class_label = row[-1]
            features = row[:-1]
            data_per_class[class_label].append(features)

        
        # Вычисляем априорные вероятности каждого класса
        self.p_classes = {
            class_label: len(features_list) / len(X)
            for class_label, features_list in data_per_class.items()
        }
        
        # Для каждого класса вычисляем оценку параметров нормального распределения (среднее и стандартное отклонение)
        self.features_distribs = {}
        for class_label, rows in data_per_class.items():
            # Преобразуем список строк в DataFrame для удобных вычислений
            data = pd.DataFrame(rows)
            means = data.mean()
            stds = data.std(ddof=1)
            # Если стандартное отклонение равно 0, заменяем его на маленькое число, чтобы избежать деления на ноль
            self.features_distribs[class_label] = {
                col: st.norm(loc=means[col], scale=stds[col] if stds[col] > 0 else 1e-6)
                for col in data.columns
            }
        
        return self
    
    def predict_prob(self, X):
        """
        Вычисляет апостериорные вероятности принадлежности к каждому классу.
        
        Args:
            X (array-like): Матрица признаков для предсказания.
            
        Returns:
            list[np.ndarray]: Список, где каждый элемент – это массив
            вероятностей принадлежности к классам для соответствующего
            входного образца. Порядок классов соответствует
            `self.p_classes.keys()`.
        """
        prob_per_class_for_x = []
        class_order = list(self.p_classes.keys())
        
        # Обходим каждый образец из выборки
        for x in X:
            total_prob = 0.0  # Сумма по классам (нормировочная константа)
            class_probs = {}
            
            # Для каждого класса вычисляем неотнормированную вероятность (априорная вероятность * произведение плотностей)
            for class_label, distributions in self.features_distribs.items():
                # Вычисляем произведение плотностей для всех признаков
                likelihood = np.prod([
                    distribution.pdf(x[int(feature_idx)]) 
                    for feature_idx, distribution in distributions.items()
                ])
                prob = self.p_classes[class_label] * likelihood
                class_probs[class_label] = prob
                total_prob += prob
            
            # Нормируем вероятности, чтобы их сумма равнялась 1
            if total_prob > 0:
                normalized_probs = {label: prob / total_prob for label, prob in class_probs.items()}
            else:
                # Если суммарное правдоподобие оказалось 0, назначаем равномерное распределение
                normalized_probs = {label: 1.0 / len(class_probs) for label in class_probs}
            
            # Собираем вероятности в массив в порядке, определённом в class_order
            prob_array = np.array([normalized_probs[label] for label in class_order])
            prob_per_class_for_x.append(prob_array)
        
        return prob_per_class_for_x
    
    def predict(self, X):
        """
        Предсказывает метки классов для входных данных.
        
        Args:
            X (array-like): Матрица признаков для предсказания.
            
        Returns:
            np.ndarray: Массив предсказанных меток классов.
        """
        # Получаем вероятностное распределение по классам для каждого образца
        probs = self.predict_prob(X)
        class_order = list(self.p_classes.keys())
        # Для каждого образца выбираем класс с максимальной вероятностью
        predictions = np.array([class_order[np.argmax(prob)] for prob in probs])
        return predictions
#
NBC = [NormalNaiveBayes]


from setuptools import setup, find_packages
import re
import os


def readme():
  with open('README.md', 'r', encoding='utf-8') as f:
    return f.read()

def get_version():
    with open(os.path.join('matplobblib', '__init__.py'), 'r') as f:
        version_file_content = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file_content, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='matplobblib',
    version=get_version(),
    packages=find_packages(),
    description='Just a library for some subjects',
    author='Ackrome',
    author_email='ivansergeyevicht@gmail.com',
    url='https://github.com/Ackrome/matplobblib',
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    include_package_data=True,  # Include non-Python files specified in MANIFEST.in or package_data
    # spackage_data=package_data,
    install_requires=[
        "numpy",
        "sympy",
        "pandas",
        "scipy",
        "pyperclip",
        "PyMuPDF",
        "graphviz",
        "statsmodels",
        "cvxopt",
        "beautifulsoup4",
        "matplotlib",
        "numba",
        "IPython",
        "tqdm",
        "scikit-learn",
        "scikit-image",
        "requests",
        "Pillow"
    ],
    license='MIT'
)



import requests
from io import BytesIO
from PIL import Image
import IPython.display as display
from ...forall import *

BASE_API_URL = r"https://api.github.com/repos/Ackrome/matplobblib/contents"
BASE_GET_URL = r"https://raw.githubusercontent.com/Ackrome/matplobblib/master"

# Список для хранения динамически созданных функций отображения теории.
####################################################################################################
def check_internet_connection(host="8.8.8.8", port=53, timeout=3):
    """
    Проверяет наличие интернет-соединения.
    """
    import socket
    try:
        socket.setdefaulttimeout(timeout)
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
        return True
    except socket.error as ex:
        return False
####################################################################################################
THEORY = []
####################################################################################################
def list_subdirectories(url=BASE_API_URL):
    """
    Получает список подкаталогов из репозитория GitHub.
    """
    if not check_internet_connection():
        print("Ошибка: Для выполнения этой функции требуется интернет-соединение.")
        return []
    
    response = requests.get(url)
    if response.status_code == 200:
        contents = response.json()
        return [item['name'] for item in contents if item['type'] == 'dir']
    else:
        print(f"Ошибка при получении подпапок: {response.status_code}")
        return []
####################################################################################################
def get_exact_format_files_from_subdir(subdir,url=BASE_API_URL,exact_format='png'):
    """
    Получает список URL-адресов .`exact_format`-файлов из указанного подкаталога в репозитории GitHub.
    """
    if not check_internet_connection():
        print("Ошибка: Для выполнения этой функции требуется интернет-соединение.")
        return []
    url = url +f"/{subdir}"
    
    response = requests.get(url)
    if response.status_code == 200:
        contents = response.json()
        exact_format_files = [item['name'] for item in contents if item['name'].endswith('.'+exact_format)]
        return [BASE_GET_URL + f"/{subdir}/{file}" for file in exact_format_files]
    else:
        print(f"Ошибка доступа к {subdir}: {response.status_code}")
        return []
####################################################################################################
def display_exact_format_files_from_subdir(subdir,url=BASE_API_URL,exact_format='png'):
    """
    Отображает exact_format-файлы из указанного подкаталога.
    """
    if not check_internet_connection():
        print("Ошибка: Для выполнения этой функции требуется интернет-соединение.")
        return
    exact_format_urls = get_exact_format_files_from_subdir(subdir,url,exact_format)
    if exact_format == 'png':
        
        for url in exact_format_urls:
            try:
                response = requests.get(url)
                response.raise_for_status()
                img = Image.open(BytesIO(response.content))
                display.display(img)
            except requests.exceptions.RequestException as e:
                print(f"Ошибка загрузки {url}: {e}")
                
    if exact_format == 'md':
        for url in exact_format_urls:
            response = requests.get(url)
            if response.status_code == 200:
                display.display(display.Markdown(response.text))
            else:
                print(f"Ошибка загрузки файла: {response.status_code}")    
####################################################################################################
# Динамическое создание функций для каждого подкаталога
def create_subdir_function(subdir,url=BASE_API_URL,exact_format='png'):
    """
    Динамически создает функцию для отображения PNG-файлов из заданного подкаталога.
    Функция именуется display_{subdir}.
    """
    global THEORY
    # Динамическое определение функции
    def display_function():
        """
        Автоматически сгенерированная функция для отображения PNG-файлов.
        """
        display_exact_format_files_from_subdir(subdir,url,exact_format)
    
    # Динамическое присвоение имени функции
    display_function.__name__ = f"display_{subdir}"
    
    # Добавление описательной строки документации
    display_function.__doc__ = (
        f"Вывести все страницы из файла с теорией '{subdir.replace('_','-')}'.\n"
        f"Эта функция сгенерирована автоматически из файла '{subdir.replace('_','-')+'.pdf'}' "
        f"из внутрибиблиотечного каталога файлов с теорией."
    )
    
    # Добавление функции в глобальное пространство имен
    globals()[display_function.__name__] = display_function
    THEORY.append(display_function)
####################################################################################################   
####################################################################################################
# По идее в pdfs должны лежать только файлы с матстатом, поэтому не стоит сильно надеяться на автосоздание этого всего
# Динамическое получение списка подкаталогов
url = BASE_API_URL+"/theory_files/pdfs"
subdirs_disp = list_subdirectories(url)
# Динамическое создание функций для каждого подкаталога
for subdir in subdirs_disp:
    if subdir.startswith('MS'):
        create_subdir_function(subdir, url)


from ..forall import *
from random import shuffle
from .data_structures import *
#######################################################################################################################
# Дополнительные функции для структур данных
#######################################################################################################################
def AISD_ADD_1(self, key):
    """
    Функция двойного хэширования для использования в хэш-таблицах с открытой адресацией.

    Args:
        self: Экземпляр класса хэш-таблицы, содержащий атрибут `size`.
        key: Ключ, для которого вычисляется вторичный хэш.

    Returns:
        int: Значение шага для пробирования, не равное нулю.
    """
    return 1 + (hash(key) % (self.size-2))
#######################################################################################################################
def AISD_ADD_2(stack):
    """
    Находит первый четный элемент в стеке, не изменяя его.

    Args:
        stack (Stack): Экземпляр класса `Stack`.

    Returns:
        int or None: Первый найденный четный элемент или None, если четных элементов нет.
    """
    current = stack.head
    while current:
        if current.data % 2 == 0:
            return current.data
        current = current.next
    return None
#######################################################################################################################
def AISD_ADD_3(stack):
    """
    Альтернативная функция для нахождения первого четного элемента в стеке.

    Эта функция изменяет стек, извлекая элементы, а затем восстанавливает его.

    Args:
        stack (Stack): Экземпляр класса `Stack`.

    Returns:
        int or None: Первый найденный четный элемент или None, если четных элементов нет.
    """
    temp_stack = Stack()
    even = None

    while not stack.is_empty():
        item = stack.pop()
        temp_stack.push(item)
        if item % 2 == 0:
            even = item
            break

    while not temp_stack.is_empty():
        stack.push(temp_stack.pop())

    return even
#######################################################################################################################
def AISD_ADD_4(stack, item):
    """
    Добавляет новый элемент в стек после первого нечетного элемента.

    Args:
        stack (Stack): Экземпляр класса `Stack`.
        item: Элемент для добавления.
    """
    current = stack.head
    while current:
        if current.data % 2 != 0:
            new_node = Node(item)
            new_node.next = current.next
            current.next = new_node
            return
        current = current.next
    stack.push(item)
#######################################################################################################################
def AISD_ADD_5(expr):
    """
    Проверяет сбалансированность скобок в математическом выражении.

    Args:
        expr (str): Строка с математическим выражением.

    Returns:
        bool: True, если скобки сбалансированы, иначе False.
    """
    stack = Stack()
    for char in expr:
        if char in "({[":
            stack.push(char)
        elif char in ")}]":
            if stack.is_empty():
                return False
            elif char == ")" and stack.peek() == "(":
                stack.pop()
            elif char == "}" and stack.peek() == "{":
                stack.pop()
            elif char == "]" and stack.peek() == "[":
                stack.pop()
            else:
                return False
    return stack.is_empty()
#######################################################################################################################
def AISD_ADD_6(expression):
    """
    Вычисляет математическое выражение в обратной польской нотации (RPN).

    Args:
        expression (list): Список токенов (числа и операторы) в RPN.

    Returns:
        float: Результат вычисления.
    """
    stack = Stack()
    for token in expression:
        if token.isdigit():
            stack.push(int(token))
        else:
            operand_2 = stack.pop()
            operand_1 = stack.pop()
            if token == '+':
                result = operand_1 + operand_2
            elif token == '-':
                result = operand_1 - operand_2
            elif token == '*':
                result = operand_1 * operand_2
            elif token == '/':
                result = operand_1 / operand_2
            stack.push(result)
    return stack.pop()
#######################################################################################################################
def AISD_ADD_7(queue):
    """
    Находит первый нечетный элемент в очереди, не изменяя ее.

    Args:
        queue (Queue): Экземпляр класса `Queue`.

    Returns:
        int or None: Первый найденный нечетный элемент или None, если таких нет.
    """
    current = queue.head
    while current:
        if current.data % 2 != 0:
            return current.data
        current = current.next
    return None
#######################################################################################################################
def AISD_ADD_8(queue, item):
    """
    Добавляет новый элемент в очередь перед первым четным элементом.

    Args:
        queue (Queue): Экземпляр класса `Queue`.
        item: Элемент для добавления.
    """
    new_node = Node(item)
    if not queue.head:
        queue.head = new_node
        queue.tail = new_node
    elif queue.head.data % 2 == 0:
        new_node.next = queue.head
        queue.head = new_node
    else:
        prev_node = queue.head
        current = prev_node.next
        while current:
            if current.data % 2 == 0:
                prev_node.next = new_node
                new_node.next = current
                return
            prev_node = current
            current = current.next
        queue.tail.next = new_node
        queue.tail = new_node
#######################################################################################################################
def AISD_ADD_9(queue, data):
    """
    Альтернативная функция для добавления нового элемента в очередь перед первым четным элементом.

    Args:
        queue (Queue): Экземпляр класса `Queue`.
        data: Элемент для добавления.
    """
    temp_queue = Queue()
    even_found = False

    while not queue.is_empty():
        item = queue.dequeue()
        if item % 2 == 0 and not even_found:
            temp_queue.enqueue(data)
            even_found = True
        temp_queue.enqueue(item)

    while not temp_queue.is_empty():
        queue.enqueue(temp_queue.dequeue())
#######################################################################################################################
def AISD_ADD_10(dllist):
    """
    Удваивает каждый четный элемент в двусвязном списке, добавляя его копию после него.

    Args:
        dllist (DoublyLinkedList): Экземпляр двусвязного списка.
    """
    current_node = dllist.head
    while current_node:
        if current_node.data % 2 == 0:
            new_node = Node(current_node.data)
            new_node.next = current_node.next
            new_node.prev = current_node
            if current_node.next:
                current_node.next.prev = new_node
            current_node.next = new_node
            current_node = new_node.next
        else:
            current_node = current_node.next
#######################################################################################################################
def AISD_ADD_11(dllist):
    """
    Удаляет все отрицательные элементы из двусвязного списка.

    Args:
        dllist (DoublyLinkedList): Экземпляр двусвязного списка.
    """
    current_node = dllist.head
    while current_node:
        if current_node.data < 0:
            if current_node.prev:
                current_node.prev.next = current_node.next
            else:
                dllist.head = current_node.next
            if current_node.next:
                current_node.next.prev = current_node.prev
        current_node = current_node.next
#######################################################################################################################
def AISD_ADD_12(cdllist):
    """
    Возводит в квадрат все отрицательные элементы в циклическом двусвязном списке.

    Args:
        cdllist (CircularDoublyLinkedList): Экземпляр циклического двусвязного списка.
    """
    current_node = cdllist.head
    while current_node:
        if current_node.data < 0:
            current_node.data = current_node.data ** 2
        current_node = current_node.next
        if current_node == cdllist.head:
            break
#######################################################################################################################
def AISD_ADD_13(cdllist):
    """
    Удаляет все элементы, кратные 5, из циклического двусвязного списка.

    Args:
        cdllist (CircularDoublyLinkedList): Экземпляр циклического двусвязного списка.
    """
    current_node = cdllist.head
    while current_node:
        if current_node.data % 5 == 0:
            cdllist.delete(current_node.data)
        current_node = current_node.next
        if current_node == cdllist.head:
            break
#######################################################################################################################
def AISD_ADD_14(levels):
    """
    Создает случайное бинарное дерево заданной глубины.

    Args:
        levels (int): Глубина создаваемого дерева.

    Returns:
        Tree: Экземпляр класса `Tree` со случайными значениями.
    """
    tree = Tree()
    nodes = 2**(levels+1) - 1
    values = list(range(1,nodes+1))
    shuffle(values)

    for i in range(nodes):
        value = values[i]
        if i == 0:
            tree.add_node(value)
        else:
            parent_index = (i-1)//2
            parent_value = values[parent_index]
            tree.add_node(value, parent_value)

    return tree
#######################################################################################################################
def AISD_ADD_15(tree, node=None):
    """
    Заменяет значение каждого узла в дереве на сумму значений всех его потомков.

    Args:
        tree (Tree): Экземпляр дерева.
        node (TreeNode, optional): Текущий узел для обработки. Defaults to `tree.root`.

    Returns:
        int: Сумма значений потомков для текущего узла.
    """
    if node is None:
        node = tree.root
    if not node.children:
        return node.value
    else:
        sum_of_children = 0
        for child in node.children:
            sum_of_children += AISD_ADD_15(tree, child)
        node.value = sum_of_children
        return sum_of_children
#######################################################################################################################
def AISD_ADD_16(tree, node=None):
    """
    Удваивает значение каждого нечетного узла в дереве.

    Args:
        tree (Tree): Экземпляр дерева.
        node (TreeNode, optional): Текущий узел для обработки. Defaults to `tree.root`.

    Returns:
        Tree: Модифицированное дерево.
    """
    if node is None:
        node = tree.root
    if node.value % 2 == 1:
        node.value *= 2
    for child in node.children:
        AISD_ADD_16(tree, child)
    return tree
#######################################################################################################################
def AISD_ADD_17(tree, node=None, leaves=None):
    """
    Находит все листья (конечные узлы) в дереве.

    Args:
        tree (Tree): Экземпляр дерева.
        node (TreeNode, optional): Текущий узел для обработки. Defaults to `tree.root`.
        leaves (list, optional): Список для сбора листьев. Defaults to None.

    Returns:
        list: Список значений всех листьев дерева.
    """
    if leaves is None:
        leaves = []
    if node is None:
        node = tree.root
    if len(node.children) == 0:
        leaves.append(node.value)
    else:
        for child in node.children:
            AISD_ADD_17(tree, child, leaves)
    return leaves
#######################################################################################################################
def AISD_ADD_18(node):
    """
    Рекурсивно находит количество узлов в бинарном дереве.

    Args:
        node (BinaryTreeNode): Корневой узел дерева или поддерева.

    Returns:
        int: Общее количество узлов.
    """
    if node is None:
        return 0
    return 1 + AISD_ADD_18(node.left) + AISD_ADD_18(node.right)
#######################################################################################################################
def AISD_ADD_19(node, target_node):
    """
    Находит путь от корня до заданного узла в бинарном дереве.

    Args:
        node (BinaryTreeNode): Текущий узел для поиска.
        target_node (BinaryTreeNode): Целевой узел.

    Returns:
        list: Список значений узлов на пути от корня до целевого узла.
    """
    if node is None:
        return []
    if node.left == target_node or node.right == target_node:
        return [node.data]
    left = AISD_ADD_19(node.left, target_node)
    right = AISD_ADD_19(node.right, target_node)
    if left:
        return [node.data] + left
    elif right:
        return [node.data] + right
    else:
        return []
#######################################################################################################################
def AISD_ADD_20(node, value):
    """
    Находит все узлы в бинарном дереве, значение которых больше или равно заданному.

    Args:
        node (BinaryTreeNode): Текущий узел для поиска.
        value: Значение для сравнения.

    Returns:
        list: Список значений узлов, удовлетворяющих условию.
    """
    if node is None:
        return []
    result = []
    if node.data >= value:
        result.append(node.data)
    result += AISD_ADD_20(node.left, value)
    result += AISD_ADD_20(node.right, value)
    return result
#######################################################################################################################
def AISD_ADD_21(hash_table):
    """
    Находит наиболее часто встречающееся значение в хэш-таблице по полю `species`.

    Args:
        hash_table (HashTableChaining): Экземпляр хэш-таблицы, где значениями являются объекты с атрибутом `species`.

    Returns:
        object: Наиболее часто встречающееся значение `species`.
    """
    species_count = {}
    for slot in hash_table.table:
        for _, animal in slot:
            if animal.species in species_count:
                species_count[animal.species] += 1
            else:
                species_count[animal.species] = 1
    return max(species_count, key=species_count.get)
#######################################################################################################################
AISD_ADD = [
    AISD_ADD_1,
    AISD_ADD_2,
    AISD_ADD_3,
    AISD_ADD_4,
    AISD_ADD_5,
    AISD_ADD_6,
    AISD_ADD_7,
    AISD_ADD_8,
    AISD_ADD_9,
    AISD_ADD_10,
    AISD_ADD_11,
    AISD_ADD_12,
    AISD_ADD_13,
    AISD_ADD_14,
    AISD_ADD_15,
    AISD_ADD_16,
    AISD_ADD_17,
    AISD_ADD_18,
    AISD_ADD_19,
    AISD_ADD_20,
    AISD_ADD_21,
]



import numpy as np
def Kfold_split(n_samples, n_splits=5, shuffle=True, random_state=None):
        """
        Генерирует индексы для разделения данных на обучающие и тестовые наборы (K-Fold).
        
        Args:
            n_samples (int): Общее количество образцов в данных.
            n_splits (int, optional): Количество фолдов (частей). Defaults to 5.
            shuffle (bool, optional): Перемешивать ли данные перед разбиением.
                Defaults to True.
            random_state (int, optional): Зерно для генератора случайных чисел
                для воспроизводимости перемешивания. Defaults to None.
          
        Yields:
            tuple[np.ndarray, np.ndarray]: Кортеж, содержащий индексы для
            обучающего и тестового наборов для каждой итерации.
        """
        indices = np.arange(n_samples)
        if shuffle:
            rng = np.random.default_rng(random_state)
            rng.shuffle(indices)
        fold_sizes = (n_samples // n_splits) * np.ones(n_splits, dtype=int)
        fold_sizes[:n_samples % n_splits] += 1
        current = 0
        for fold_size in fold_sizes:
            test_indices = indices[current:current + fold_size]
            train_indices = np.concatenate([indices[:current], indices[current + fold_size:]])
            yield train_indices, test_indices
            current += fold_size

FOLDS = [Kfold_split]


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
#######################################################################################################################
# Дополнительные функции
#######################################################################################################################
def B2_VISUAL(b, k = 1,lamb_l1=0,lamb_l2=0,num_points=100):
  """
  Визуализирует траекторию изменения весов модели и изолинии L1/L2 регуляризации.

  Args:
      b (pd.DataFrame): DataFrame с двумя столбцами, представляющими
          изменение весов модели на каждой итерации.
      k (int, optional): Масштаб для расширения границ графика. Defaults to 1.
      lamb_l1 (float, optional): Параметр L1 регуляризации (λ). Если не 0,
          будут построены изолинии L1. Defaults to 0.
      lamb_l2 (float, optional): Параметр L2 регуляризации (λ). Если не 0,
          будут построены изолинии L2. Defaults to 0.
      num_points (int, optional): Количество точек для построения сетки
          изолиний. Defaults to 100.
  """
  plt.plot(b[0],b[1])
  plt.grid('ON')
  plt.xlim(-np.max(b) - k,np.max(b) + k)
  plt.ylim(-np.max(b) - k,np.max(b) + k)

  limit = np.max(np.abs(b)) + 1
  # Create a grid of b1 and b2 values
  b1 = np.linspace(-limit, limit, num_points)
  b2 = np.linspace(-limit, limit, num_points)
  B1, B2 = np.meshgrid(b1, b2)

  # Calculate L1 and L2 penalties
  if (lamb_l1!=0):
    L1_penalty = lamb_l1 * (np.abs(B1) + np.abs(B2))
    # Plot L1 regularization isolines with gradient color change
    contour_L1 = plt.contourf(B1, B2, L1_penalty, levels=20, cmap="Oranges_r")
    plt.colorbar(contour_L1, label="L1 Penalty")
    plt.plot(b[0], b[1], "bo-", label="Trajectory")
    plt.title("L1 Regularization Isolines")
    plt.xlabel("$b_1$")
    plt.ylabel("$b_2$")
    plt.grid(True)
    plt.legend()
    
  if (lamb_l2!=0):
    L2_penalty = lamb_l2 * (B1**2 + B2**2)
    contour_L2 = plt.contourf(B1, B2, L2_penalty, levels=20, cmap="Blues_r")
    plt.colorbar(contour_L2, label="L2 Penalty")
    plt.plot(b[0], b[1], "bo-", label="Trajectory")
    plt.title("L2 Regularization Isolines")
    plt.xlabel("$b_1$")
    plt.ylabel("$b_2$")
    plt.grid(True)


  plt.show()
#######################################################################################################################
B2 = [B2_VISUAL]


import pytest

def test_package_import():
    """Tests that the main package can be imported."""
    try:
        import matplobblib
    except ImportError as e:
        pytest.fail(f"Failed to import the main matplobblib package: {e}")

def test_submodule_imports():
    """Tests that all major submodules can be imported."""
    submodules = ['aisd', 'ml', 'nm', 'tvims','tod']
    for module in submodules:
        try:
            # Dynamically import submodule from matplobblib
            __import__(f"matplobblib.{module}")
        except ImportError as e:
            pytest.fail(f"Failed to import submodule matplobblib.{module}: {e}")
